from typing import Any, Dict, List, Set, Union, cast

from ..language import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    InlineFragmentNode,
    SelectionSetNode,
)
from ..type import (
    GraphQLAbstractType,
    GraphQLIncludeDirective,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLSkipDirective,
    is_abstract_type,
)
from ..utilities.type_from_ast import type_from_ast
from .values import get_directive_values

__all__ = ["collect_fields", "collect_sub_fields"]


def collect_fields(
    schema: GraphQLSchema,
    fragments: Dict[str, FragmentDefinitionNode],
    variable_values: Dict[str, Any],
    runtime_type: GraphQLObjectType,
    selection_set: SelectionSetNode,
) -> Dict[str, List[FieldNode]]:
    """Collect fields.

    Given a selection_set, collects all the fields and returns them.

    collect_fields requires the "runtime type" of an object. For a field that
    returns an Interface or Union type, the "runtime type" will be the actual
    object type returned by that field.

    For internal use only.
    """
    fields: Dict[str, List[FieldNode]] = {}
    collect_fields_impl(
        schema, fragments, variable_values, runtime_type, selection_set, fields, set()
    )
    return fields


def collect_sub_fields(
    schema: GraphQLSchema,
    fragments: Dict[str, FragmentDefinitionNode],
    variable_values: Dict[str, Any],
    return_type: GraphQLObjectType,
    field_nodes: List[FieldNode],
) -> Dict[str, List[FieldNode]]:
    """Collect sub fields.

    Given a list of field nodes, collects all the subfields of the passed in fields,
    and returns them at the end.

    collect_sub_fields requires the "return type" of an object. For a field that
    returns an Interface or Union type, the "return type" will be the actual
    object type returned by that field.

    For internal use only.
    """
    sub_field_nodes: Dict[str, List[FieldNode]] = {}
    visited_fragment_names: Set[str] = set()
    for node in field_nodes:
        if node.selection_set:
            collect_fields_impl(
                schema,
                fragments,
                variable_values,
                return_type,
                node.selection_set,
                sub_field_nodes,
                visited_fragment_names,
            )
    return sub_field_nodes


def collect_fields_impl(
    schema: GraphQLSchema,
    fragments: Dict[str, FragmentDefinitionNode],
    variable_values: Dict[str, Any],
    runtime_type: GraphQLObjectType,
    selection_set: SelectionSetNode,
    fields: Dict[str, List[FieldNode]],
    visited_fragment_names: Set[str],
) -> None:
    """Collect fields (internal implementation)."""
    for selection in selection_set.selections:
        if isinstance(selection, FieldNode):
            if not should_include_node(variable_values, selection):
                continue
            name = get_field_entry_key(selection)
            fields.setdefault(name, []).append(selection)
        elif isinstance(selection, InlineFragmentNode):
            if not should_include_node(
                variable_values, selection
            ) or not does_fragment_condition_match(schema, selection, runtime_type):
                continue
            collect_fields_impl(
                schema,
                fragments,
                variable_values,
                runtime_type,
                selection.selection_set,
                fields,
                visited_fragment_names,
            )
        elif isinstance(selection, FragmentSpreadNode):  # pragma: no cover else
            frag_name = selection.name.value
            if frag_name in visited_fragment_names or not should_include_node(
                variable_values, selection
            ):
                continue
            visited_fragment_names.add(frag_name)
            fragment = fragments.get(frag_name)
            if not fragment or not does_fragment_condition_match(
                schema, fragment, runtime_type
            ):
                continue
            collect_fields_impl(
                schema,
                fragments,
                variable_values,
                runtime_type,
                fragment.selection_set,
                fields,
                visited_fragment_names,
            )


def should_include_node(
    variable_values: Dict[str, Any],
    node: Union[FragmentSpreadNode, FieldNode, InlineFragmentNode],
) -> bool:
    """Check if node should be included

    Determines if a field should be included based on the @include and @skip
    directives, where @skip has higher precedence than @include.
    """
    skip = get_directive_values(GraphQLSkipDirective, node, variable_values)
    if skip and skip["if"]:
        return False

    include = get_directive_values(GraphQLIncludeDirective, node, variable_values)
    if include and not include["if"]:
        return False

    return True


def does_fragment_condition_match(
    schema: GraphQLSchema,
    fragment: Union[FragmentDefinitionNode, InlineFragmentNode],
    type_: GraphQLObjectType,
) -> bool:
    """Determine if a fragment is applicable to the given type."""
    type_condition_node = fragment.type_condition
    if not type_condition_node:
        return True
    conditional_type = type_from_ast(schema, type_condition_node)
    if conditional_type is type_:
        return True
    if is_abstract_type(conditional_type):
        return schema.is_sub_type(cast(GraphQLAbstractType, conditional_type), type_)
    return False


def get_field_entry_key(node: FieldNode) -> str:
    """Implements the logic to compute the key of a given field's entry"""
    return node.alias.value if node.alias else node.name.value



from copy import copy
from functools import partial
from typing import cast, List, Optional

from pytest import mark, raises

from graphql.language import (
    Node,
    FieldNode,
    NameNode,
    SelectionNode,
    SelectionSetNode,
    parse,
    visit,
    BREAK,
    REMOVE,
    SKIP,
    ParallelVisitor,
    Visitor,
    VisitorKeyMap,
)

from ..fixtures import kitchen_sink_query  # noqa: F401


def check_visitor_fn_args(ast, node, key, parent, path, ancestors, is_edited=False):
    assert isinstance(node, Node)

    is_root = key is None
    if is_root:
        if not is_edited:
            assert node is ast
        assert parent is None
        assert path == []
        assert ancestors == []
        return

    assert isinstance(key, (int, str))

    if isinstance(key, int):
        assert isinstance(parent, tuple)
        assert 0 <= key <= len(parent)
    else:
        assert isinstance(parent, Node)
        assert hasattr(parent, key)

    assert isinstance(path, list)
    assert path[-1] == key

    assert isinstance(ancestors, list)
    assert len(ancestors) == len(path) - 1

    if not is_edited:
        current_node = ast

        for i, ancestor in enumerate(ancestors):
            assert ancestor is current_node
            k = path[i]
            assert isinstance(k, (int, str))
            if isinstance(k, int):
                assert isinstance(current_node, tuple)
                assert 0 <= k <= len(current_node)
                current_node = current_node[k]
            else:
                assert isinstance(current_node, Node)
                assert hasattr(current_node, k)
                current_node = getattr(current_node, k)
            assert current_node is not None

        assert parent is current_node
        if isinstance(key, int):
            assert parent[key] is node
        else:
            assert getattr(parent, key) is node


check_visitor_fn_args_edited = partial(check_visitor_fn_args, is_edited=True)


def get_value(node):
    return getattr(node, "value", None)


def describe_visitor():
    def visit_with_invalid_node():
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            visit("invalid", Visitor())  # type: ignore
        assert str(exc_info.value) == "Not an AST Node: 'invalid'."

    def visit_with_invalid_visitor():
        ast = parse("{ a }", no_location=True)

        class TestVisitor:
            def enter(self, *_args):
                pass

        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            visit(ast, TestVisitor())  # type: ignore
        assert str(exc_info.value) == "Not an AST Visitor: <TestVisitor instance>."

    def visitors_support_all_method_variants():
        class TestVisitorWithInstanceMethods(Visitor):
            def enter(self, node, *args):
                assert isinstance(self, TestVisitorWithInstanceMethods)
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"enter:{node.kind}")
                pass

            def leave(self, node, *args):
                assert isinstance(self, TestVisitorWithInstanceMethods)
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"leave:{node.kind}")
                pass

            def enter_field(self, node, *args):
                assert isinstance(self, TestVisitorWithInstanceMethods)
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"enter_field:{node.kind}")
                pass

            def leave_field(self, node, *args):
                assert isinstance(self, TestVisitorWithInstanceMethods)
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"leave_field:{node.kind}")
                pass

        class TestVisitorWithClassMethods(Visitor):
            @classmethod
            def enter(cls, node, *args):
                assert cls is TestVisitorWithClassMethods
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"enter:{node.kind}")
                pass

            @classmethod
            def leave(cls, node, *args):
                assert cls is TestVisitorWithClassMethods
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"leave:{node.kind}")
                pass

            @classmethod
            def enter_field(cls, node, *args):
                assert cls is TestVisitorWithClassMethods
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"enter_field:{node.kind}")
                pass

            @classmethod
            def leave_field(cls, node, *args):
                assert cls is TestVisitorWithClassMethods
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"leave_field:{node.kind}")
                pass

        class TestVisitorWithStaticMethods(Visitor):
            @staticmethod
            def enter(node, *args):
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"enter:{node.kind}")
                pass

            @staticmethod
            def leave(node, *args):
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"leave:{node.kind}")
                pass

            @staticmethod
            def enter_field(node, *args):
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"enter_field:{node.kind}")
                pass

            @staticmethod
            def leave_field(node, *args):
                assert isinstance(node, Node)
                assert len(args) == 4
                visited.append(f"leave_field:{node.kind}")
                pass

        for visitor_class in (
            TestVisitorWithInstanceMethods,
            TestVisitorWithClassMethods,
            TestVisitorWithStaticMethods,
        ):
            ast = parse("{ a }")
            visited: List[str] = []
            visit(ast, visitor_class())
            assert visited == [
                "enter:document",
                "enter:operation_definition",
                "enter:selection_set",
                "enter_field:field",
                "enter:name",
                "leave:name",
                "leave_field:field",
                "leave:selection_set",
                "leave:operation_definition",
                "leave:document",
            ]

    def has_get_enter_leave_for_kind_method():
        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                pass

            @staticmethod
            def enter_document(*args):
                pass

            @staticmethod
            def leave(*args):
                pass

            @staticmethod
            def leave_document(*args):
                pass

        visitor = TestVisitor()

        assert visitor.get_enter_leave_for_kind("document") == (
            visitor.enter_document,
            visitor.leave_document,
        )
        assert visitor.get_enter_leave_for_kind("field") == (
            visitor.enter,
            visitor.leave,
        )

        # also test deprecated method
        assert visitor.get_visit_fn("document") == visitor.enter_document
        assert visitor.get_visit_fn("field") == visitor.enter
        assert visitor.get_visit_fn("document", True) == visitor.leave_document
        assert visitor.get_visit_fn("field", True) == visitor.leave

    def validates_path_argument():
        ast = parse("{ a }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                visited.append(["enter", *args[3]])

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                visited.append(["leave", *args[3]])

        visit(ast, TestVisitor())
        assert visited == [
            ["enter"],
            ["enter", "definitions", 0],
            ["enter", "definitions", 0, "selection_set"],
            ["enter", "definitions", 0, "selection_set", "selections", 0],
            ["enter", "definitions", 0, "selection_set", "selections", 0, "name"],
            ["leave", "definitions", 0, "selection_set", "selections", 0, "name"],
            ["leave", "definitions", 0, "selection_set", "selections", 0],
            ["leave", "definitions", 0, "selection_set"],
            ["leave", "definitions", 0],
            ["leave"],
        ]

    def validates_ancestors_argument():
        ast = parse("{ a }", no_location=True)
        visited_nodes = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(node, key, parent, _path, ancestors):
                in_array = isinstance(key, int)
                if in_array:
                    visited_nodes.append(parent)
                visited_nodes.append(node)
                expected_ancestors = visited_nodes[0:-2]
                assert ancestors == expected_ancestors

            @staticmethod
            def leave(_node, key, _parent, _path, ancestors):
                expected_ancestors = visited_nodes[0:-2]
                assert ancestors == expected_ancestors
                in_array = isinstance(key, int)
                if in_array:
                    visited_nodes.pop()
                visited_nodes.pop()

        visit(ast, TestVisitor())

    def allows_visiting_only_specified_nodes():
        ast = parse("{ a }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            selection_set = None

            @staticmethod
            def enter_field(node, *_args):
                visited.append(["enter", node.kind])

            @staticmethod
            def leave_field(node, *_args):
                visited.append(["leave", node.kind])

        visit(ast, TestVisitor())
        assert visited == [["enter", "field"], ["leave", "field"]]

    def allows_editing_a_node_both_on_enter_and_on_leave():
        ast = parse("{ a, b, c { a, b, c } }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            selection_set = None

            def enter_operation_definition(self, *args):
                check_visitor_fn_args(ast, *args)
                node = copy(args[0])
                assert len(node.selection_set.selections) == 3
                self.selection_set = node.selection_set
                node.selection_set = SelectionSetNode(selections=[])
                visited.append("enter")
                return node

            def leave_operation_definition(self, *args):
                check_visitor_fn_args_edited(ast, *args)
                node = copy(args[0])
                assert not node.selection_set.selections
                node.selection_set = self.selection_set
                visited.append("leave")
                return node

        edited_ast = visit(ast, TestVisitor())
        assert edited_ast == ast
        assert visited == ["enter", "leave"]

    @mark.parametrize("remove_action", (REMOVE, Ellipsis), ids=("REMOVE", "Ellipsis"))
    def allows_for_editing_on_enter(remove_action):
        ast = parse("{ a, b, c { a, b, c } }", no_location=True)

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                if isinstance(node, FieldNode) and node.name.value == "b":
                    return remove_action

        edited_ast = visit(ast, TestVisitor())
        assert ast == parse("{ a, b, c { a, b, c } }", no_location=True)
        assert edited_ast == parse("{ a,    c { a,    c } }", no_location=True)

    @mark.parametrize("remove_action", (REMOVE, Ellipsis), ids=("REMOVE", "Ellipsis"))
    def allows_for_editing_on_leave(remove_action):
        ast = parse("{ a, b, c { a, b, c } }", no_location=True)

        class TestVisitor(Visitor):
            @staticmethod
            def leave(*args):
                check_visitor_fn_args_edited(ast, *args)
                node = args[0]
                if isinstance(node, FieldNode) and node.name.value == "b":
                    return remove_action

        edited_ast = visit(ast, TestVisitor())
        assert ast == parse("{ a, b, c { a, b, c } }", no_location=True)
        assert edited_ast == parse("{ a,    c { a,    c } }", no_location=True)

    @mark.parametrize("skip_action", (SKIP, False), ids=("SKIP", "False"))
    def ignores_false_returned_on_leave(skip_action):
        ast = parse("{ a, b, c { a, b, c } }", no_location=True)

        class TestVisitor(Visitor):
            @staticmethod
            def leave(*args):
                return skip_action

        returned_ast = visit(ast, TestVisitor())
        assert returned_ast == parse("{ a, b, c { a, b, c } }", no_location=True)

    def visits_edited_node():
        ast = parse("{ a { x } }", no_location=True)
        added_field = FieldNode(name=NameNode(value="__typename"))

        class TestVisitor(Visitor):
            did_visit_added_field = False

            def enter(self, *args):
                check_visitor_fn_args_edited(ast, *args)
                node = args[0]
                if isinstance(node, FieldNode) and node.name.value == "a":
                    node = copy(node)
                    assert node.selection_set
                    node.selection_set.selections = (
                        added_field,
                    ) + node.selection_set.selections
                    return node
                if node == added_field:
                    self.did_visit_added_field = True

        visitor = TestVisitor()
        visit(ast, visitor)
        assert visitor.did_visit_added_field

    @mark.parametrize("skip_action", (SKIP, False), ids=("SKIP", "False"))
    def allows_skipping_a_sub_tree(skip_action):
        ast = parse("{ a, b { x }, c }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])
                if kind == "field" and node.name.value == "b":
                    return skip_action

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        visit(ast, TestVisitor())
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "field", None],
            ["enter", "name", "c"],
            ["leave", "name", "c"],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "operation_definition", None],
            ["leave", "document", None],
        ]

    @mark.parametrize("break_action", (BREAK, True), ids=("BREAK", "True"))
    def allows_early_exit_while_visiting(break_action):
        ast = parse("{ a, b { x }, c }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])
                if kind == "name" and node.value == "x":
                    return break_action

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        visit(ast, TestVisitor())
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "b"],
            ["leave", "name", "b"],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "x"],
        ]

    @mark.parametrize("break_action", (BREAK, True), ids=("BREAK", "True"))
    def allows_early_exit_while_leaving(break_action):
        ast = parse("{ a, b { x }, c }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])
                if kind == "name" and node.value == "x":
                    return break_action

        visit(ast, TestVisitor())
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "b"],
            ["leave", "name", "b"],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "x"],
            ["leave", "name", "x"],
        ]

    def allows_a_named_functions_visitor_api():
        ast = parse("{ a, b { x }, c }", no_location=True)
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter_name(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def enter_selection_set(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def leave_selection_set(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        visit(ast, TestVisitor())
        assert visited == [
            ["enter", "selection_set", None],
            ["enter", "name", "a"],
            ["enter", "name", "b"],
            ["enter", "selection_set", None],
            ["enter", "name", "x"],
            ["leave", "selection_set", None],
            ["enter", "name", "c"],
            ["leave", "selection_set", None],
        ]

    def visit_nodes_with_custom_kinds_but_does_not_traverse_deeper():
        # GraphQL.js removed support for unknown node types,
        # but it is easy for us to add and support custom node types,
        # so we keep allowing this and test this feature here.
        custom_ast = parse("{ a }")

        class CustomFieldNode(SelectionNode):
            __slots__ = "name", "selection_set"

            name: NameNode
            selection_set: Optional[SelectionSetNode]

        custom_selection_set = cast(FieldNode, custom_ast.definitions[0]).selection_set
        assert custom_selection_set is not None
        custom_selection_set.selections = custom_selection_set.selections + (
            CustomFieldNode(
                name=NameNode(value="NameNodeToBeSkipped"),
                selection_set=SelectionSetNode(
                    selections=CustomFieldNode(
                        name=NameNode(value="NameNodeToBeSkipped")
                    )
                ),
            ),
        )

        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(node, *_args):
                visited.append(["enter", node.kind, get_value(node)])

            @staticmethod
            def leave(node, *_args):
                visited.append(["leave", node.kind, get_value(node)])

        visit(custom_ast, TestVisitor())
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "custom_field", None],
            ["leave", "custom_field", None],
            ["leave", "selection_set", None],
            ["leave", "operation_definition", None],
            ["leave", "document", None],
        ]

    def visits_only_the_specified_kind_in_visitor_key_map():
        visited = []

        visitor_key_map: VisitorKeyMap = {
            "document": ("definitions",),
            "operation_definition": ("name",),
        }

        class TestVisitor(Visitor):
            @staticmethod
            def enter(node, *_args):
                visited.append(["enter", node.kind, get_value(node)])

            @staticmethod
            def leave(node, *_args):
                visited.append(["leave", node.kind, get_value(node)])

        example_document_ast = parse(
            """
            query ExampleOperation {
              someField
            }
            """
        )

        visit(example_document_ast, TestVisitor(), visitor_key_map)
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "name", "ExampleOperation"],
            ["leave", "name", "ExampleOperation"],
            ["leave", "operation_definition", None],
            ["leave", "document", None],
        ]

    def cannot_define_visitor_with_unknown_ast_nodes():
        with raises(TypeError) as exc_info:

            class VisitorWithNonExistingNode(Visitor):
                def enter_field(self, *_args):
                    pass

                def leave_garfield(self, *_args):
                    pass

        assert str(exc_info.value) == "Invalid AST node kind: garfield."

        with raises(TypeError) as exc_info:

            class VisitorWithUnspecificNode(Visitor):
                def enter_type_system_extension(self, *_args):
                    pass

        assert str(exc_info.value) == "Invalid AST node kind: type_system_extension."

    def legacy_visits_variables_defined_in_fragments():
        ast = parse(
            "fragment a($v: Boolean = false) on t { f }",
            no_location=True,
            allow_legacy_fragment_variables=True,
        )
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        visit(ast, TestVisitor())
        assert visited == [
            ["enter", "document", None],
            ["enter", "fragment_definition", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["enter", "variable_definition", None],
            ["enter", "variable", None],
            ["enter", "name", "v"],
            ["leave", "name", "v"],
            ["leave", "variable", None],
            ["enter", "named_type", None],
            ["enter", "name", "Boolean"],
            ["leave", "name", "Boolean"],
            ["leave", "named_type", None],
            ["enter", "boolean_value", False],
            ["leave", "boolean_value", False],
            ["leave", "variable_definition", None],
            ["enter", "named_type", None],
            ["enter", "name", "t"],
            ["leave", "name", "t"],
            ["leave", "named_type", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "f"],
            ["leave", "name", "f"],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "fragment_definition", None],
            ["leave", "document", None],
        ]

    # noinspection PyShadowingNames
    def visits_kitchen_sink(kitchen_sink_query):  # noqa: F811
        ast = parse(kitchen_sink_query)
        visited: List = []
        record = visited.append
        arg_stack: List = []
        push = arg_stack.append
        pop = arg_stack.pop

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                node, key, parent = args[:3]
                parent_kind = parent.kind if isinstance(parent, Node) else None
                record(["enter", node.kind, key, parent_kind])

                check_visitor_fn_args(ast, *args)
                push(args[:])

            @staticmethod
            def leave(*args):
                node, key, parent = args[:3]
                parent_kind = parent.kind if isinstance(parent, Node) else None
                record(["leave", node.kind, key, parent_kind])

                assert pop() == args

        visit(ast, TestVisitor())

        assert arg_stack == []
        assert visited == [
            ["enter", "document", None, None],
            ["enter", "operation_definition", 0, None],
            ["enter", "name", "name", "operation_definition"],
            ["leave", "name", "name", "operation_definition"],
            ["enter", "variable_definition", 0, None],
            ["enter", "variable", "variable", "variable_definition"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "variable", "variable_definition"],
            ["enter", "named_type", "type", "variable_definition"],
            ["enter", "name", "name", "named_type"],
            ["leave", "name", "name", "named_type"],
            ["leave", "named_type", "type", "variable_definition"],
            ["leave", "variable_definition", 0, None],
            ["enter", "variable_definition", 1, None],
            ["enter", "variable", "variable", "variable_definition"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "variable", "variable_definition"],
            ["enter", "named_type", "type", "variable_definition"],
            ["enter", "name", "name", "named_type"],
            ["leave", "name", "name", "named_type"],
            ["leave", "named_type", "type", "variable_definition"],
            ["enter", "enum_value", "default_value", "variable_definition"],
            ["leave", "enum_value", "default_value", "variable_definition"],
            ["leave", "variable_definition", 1, None],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "operation_definition"],
            ["enter", "field", 0, None],
            ["enter", "name", "alias", "field"],
            ["leave", "name", "alias", "field"],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "list_value", "value", "argument"],
            ["enter", "int_value", 0, None],
            ["leave", "int_value", 0, None],
            ["enter", "int_value", 1, None],
            ["leave", "int_value", 1, None],
            ["leave", "list_value", "value", "argument"],
            ["leave", "argument", 0, None],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["enter", "inline_fragment", 1, None],
            ["enter", "named_type", "type_condition", "inline_fragment"],
            ["enter", "name", "name", "named_type"],
            ["leave", "name", "name", "named_type"],
            ["leave", "named_type", "type_condition", "inline_fragment"],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "inline_fragment"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["enter", "field", 1, None],
            ["enter", "name", "alias", "field"],
            ["leave", "name", "alias", "field"],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "int_value", "value", "argument"],
            ["leave", "int_value", "value", "argument"],
            ["leave", "argument", 0, None],
            ["enter", "argument", 1, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "variable", "value", "argument"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "value", "argument"],
            ["leave", "argument", 1, None],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "variable", "value", "argument"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "value", "argument"],
            ["leave", "argument", 0, None],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["enter", "fragment_spread", 1, None],
            ["enter", "name", "name", "fragment_spread"],
            ["leave", "name", "name", "fragment_spread"],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["leave", "fragment_spread", 1, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 1, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "inline_fragment"],
            ["leave", "inline_fragment", 1, None],
            ["enter", "inline_fragment", 2, None],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "variable", "value", "argument"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "value", "argument"],
            ["leave", "argument", 0, None],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "inline_fragment"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "inline_fragment"],
            ["leave", "inline_fragment", 2, None],
            ["enter", "inline_fragment", 3, None],
            ["enter", "selection_set", "selection_set", "inline_fragment"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "inline_fragment"],
            ["leave", "inline_fragment", 3, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "operation_definition"],
            ["leave", "operation_definition", 0, None],
            ["enter", "operation_definition", 1, None],
            ["enter", "name", "name", "operation_definition"],
            ["leave", "name", "name", "operation_definition"],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "operation_definition"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "int_value", "value", "argument"],
            ["leave", "int_value", "value", "argument"],
            ["leave", "argument", 0, None],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "operation_definition"],
            ["leave", "operation_definition", 1, None],
            ["enter", "operation_definition", 2, None],
            ["enter", "name", "name", "operation_definition"],
            ["leave", "name", "name", "operation_definition"],
            ["enter", "variable_definition", 0, None],
            ["enter", "variable", "variable", "variable_definition"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "variable", "variable_definition"],
            ["enter", "named_type", "type", "variable_definition"],
            ["enter", "name", "name", "named_type"],
            ["leave", "name", "name", "named_type"],
            ["leave", "named_type", "type", "variable_definition"],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["leave", "variable_definition", 0, None],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "operation_definition"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "variable", "value", "argument"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "value", "argument"],
            ["leave", "argument", 0, None],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["enter", "field", 1, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "selection_set", "selection_set", "field"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 1, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "operation_definition"],
            ["leave", "operation_definition", 2, None],
            ["enter", "fragment_definition", 3, None],
            ["enter", "name", "name", "fragment_definition"],
            ["leave", "name", "name", "fragment_definition"],
            ["enter", "named_type", "type_condition", "fragment_definition"],
            ["enter", "name", "name", "named_type"],
            ["leave", "name", "name", "named_type"],
            ["leave", "named_type", "type_condition", "fragment_definition"],
            ["enter", "directive", 0, None],
            ["enter", "name", "name", "directive"],
            ["leave", "name", "name", "directive"],
            ["leave", "directive", 0, None],
            ["enter", "selection_set", "selection_set", "fragment_definition"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "variable", "value", "argument"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "value", "argument"],
            ["leave", "argument", 0, None],
            ["enter", "argument", 1, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "variable", "value", "argument"],
            ["enter", "name", "name", "variable"],
            ["leave", "name", "name", "variable"],
            ["leave", "variable", "value", "argument"],
            ["leave", "argument", 1, None],
            ["enter", "argument", 2, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "object_value", "value", "argument"],
            ["enter", "object_field", 0, None],
            ["enter", "name", "name", "object_field"],
            ["leave", "name", "name", "object_field"],
            ["enter", "string_value", "value", "object_field"],
            ["leave", "string_value", "value", "object_field"],
            ["leave", "object_field", 0, None],
            ["enter", "object_field", 1, None],
            ["enter", "name", "name", "object_field"],
            ["leave", "name", "name", "object_field"],
            ["enter", "string_value", "value", "object_field"],
            ["leave", "string_value", "value", "object_field"],
            ["leave", "object_field", 1, None],
            ["leave", "object_value", "value", "argument"],
            ["leave", "argument", 2, None],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "fragment_definition"],
            ["leave", "fragment_definition", 3, None],
            ["enter", "operation_definition", 4, None],
            ["enter", "selection_set", "selection_set", "operation_definition"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["enter", "argument", 0, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "boolean_value", "value", "argument"],
            ["leave", "boolean_value", "value", "argument"],
            ["leave", "argument", 0, None],
            ["enter", "argument", 1, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "boolean_value", "value", "argument"],
            ["leave", "boolean_value", "value", "argument"],
            ["leave", "argument", 1, None],
            ["enter", "argument", 2, None],
            ["enter", "name", "name", "argument"],
            ["leave", "name", "name", "argument"],
            ["enter", "null_value", "value", "argument"],
            ["leave", "null_value", "value", "argument"],
            ["leave", "argument", 2, None],
            ["leave", "field", 0, None],
            ["enter", "field", 1, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 1, None],
            ["leave", "selection_set", "selection_set", "operation_definition"],
            ["leave", "operation_definition", 4, None],
            ["enter", "operation_definition", 5, None],
            ["enter", "selection_set", "selection_set", "operation_definition"],
            ["enter", "field", 0, None],
            ["enter", "name", "name", "field"],
            ["leave", "name", "name", "field"],
            ["leave", "field", 0, None],
            ["leave", "selection_set", "selection_set", "operation_definition"],
            ["leave", "operation_definition", 5, None],
            ["leave", "document", None, None],
        ]


def describe_visit_in_parallel():
    @mark.parametrize("skip_action", (SKIP, False), ids=("SKIP", "False"))
    def allows_skipping_a_sub_tree(skip_action):
        # Note: nearly identical to the above test but using ParallelVisitor
        ast = parse("{ a, b { x }, c }")
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])
                if kind == "field" and node.name.value == "b":
                    return skip_action

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        visit(ast, ParallelVisitor([TestVisitor()]))
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "field", None],
            ["enter", "name", "c"],
            ["leave", "name", "c"],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "operation_definition", None],
            ["leave", "document", None],
        ]

    @mark.parametrize("skip_action", (SKIP, False), ids=("SKIP", "False"))
    def allows_skipping_different_sub_trees(skip_action):
        ast = parse("{ a { x }, b { y} }")
        visited = []

        class TestVisitor(Visitor):
            def __init__(self, name):
                super().__init__()
                self.name = name

            def enter(self, *args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                name = self.name
                visited.append([f"no-{name}", "enter", kind, value])
                if kind == "field" and node.name.value == name:
                    return skip_action

            def leave(self, *args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                name = self.name
                visited.append([f"no-{name}", "leave", kind, value])

        visit(ast, ParallelVisitor([TestVisitor("a"), TestVisitor("b")]))
        assert visited == [
            ["no-a", "enter", "document", None],
            ["no-b", "enter", "document", None],
            ["no-a", "enter", "operation_definition", None],
            ["no-b", "enter", "operation_definition", None],
            ["no-a", "enter", "selection_set", None],
            ["no-b", "enter", "selection_set", None],
            ["no-a", "enter", "field", None],
            ["no-b", "enter", "field", None],
            ["no-b", "enter", "name", "a"],
            ["no-b", "leave", "name", "a"],
            ["no-b", "enter", "selection_set", None],
            ["no-b", "enter", "field", None],
            ["no-b", "enter", "name", "x"],
            ["no-b", "leave", "name", "x"],
            ["no-b", "leave", "field", None],
            ["no-b", "leave", "selection_set", None],
            ["no-b", "leave", "field", None],
            ["no-a", "enter", "field", None],
            ["no-b", "enter", "field", None],
            ["no-a", "enter", "name", "b"],
            ["no-a", "leave", "name", "b"],
            ["no-a", "enter", "selection_set", None],
            ["no-a", "enter", "field", None],
            ["no-a", "enter", "name", "y"],
            ["no-a", "leave", "name", "y"],
            ["no-a", "leave", "field", None],
            ["no-a", "leave", "selection_set", None],
            ["no-a", "leave", "field", None],
            ["no-a", "leave", "selection_set", None],
            ["no-b", "leave", "selection_set", None],
            ["no-a", "leave", "operation_definition", None],
            ["no-b", "leave", "operation_definition", None],
            ["no-a", "leave", "document", None],
            ["no-b", "leave", "document", None],
        ]

    @mark.parametrize("break_action", (BREAK, True), ids=("BREAK", "True"))
    def allows_early_exit_while_visiting(break_action):
        # Note: nearly identical to the above test but using ParallelVisitor.
        ast = parse("{ a, b { x }, c }")
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])
                if kind == "name" and node.value == "x":
                    return break_action

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        visit(ast, ParallelVisitor([TestVisitor()]))
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "b"],
            ["leave", "name", "b"],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "x"],
        ]

    @mark.parametrize("break_action", (BREAK, True), ids=("BREAK", "True"))
    def allows_early_exit_from_different_points(break_action):
        ast = parse("{ a { y }, b { x } }")
        visited = []

        class TestVisitor(Visitor):
            def __init__(self, name):
                super().__init__()
                self.name = name

            def enter(self, *args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                name = self.name
                visited.append([f"break-{name}", "enter", kind, value])
                if kind == "name" and node.value == name:
                    return break_action

            def leave(self, *args):
                assert self.name == "b"
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                name = self.name
                visited.append([f"break-{name}", "leave", kind, value])

        visit(ast, ParallelVisitor([TestVisitor("a"), TestVisitor("b")]))
        assert visited == [
            ["break-a", "enter", "document", None],
            ["break-b", "enter", "document", None],
            ["break-a", "enter", "operation_definition", None],
            ["break-b", "enter", "operation_definition", None],
            ["break-a", "enter", "selection_set", None],
            ["break-b", "enter", "selection_set", None],
            ["break-a", "enter", "field", None],
            ["break-b", "enter", "field", None],
            ["break-a", "enter", "name", "a"],
            ["break-b", "enter", "name", "a"],
            ["break-b", "leave", "name", "a"],
            ["break-b", "enter", "selection_set", None],
            ["break-b", "enter", "field", None],
            ["break-b", "enter", "name", "y"],
            ["break-b", "leave", "name", "y"],
            ["break-b", "leave", "field", None],
            ["break-b", "leave", "selection_set", None],
            ["break-b", "leave", "field", None],
            ["break-b", "enter", "field", None],
            ["break-b", "enter", "name", "b"],
        ]

    @mark.parametrize("break_action", (BREAK, True), ids=("BREAK", "True"))
    def allows_early_exit_while_leaving(break_action):
        # Note: nearly identical to the above test but using ParallelVisitor.
        ast = parse("{ a, b { x }, c }")
        visited = []

        class TestVisitor(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def leave(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])
                if kind == "name" and node.value == "x":
                    return break_action

        visit(ast, ParallelVisitor([TestVisitor()]))
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "b"],
            ["leave", "name", "b"],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "x"],
            ["leave", "name", "x"],
        ]

    @mark.parametrize("break_action", (BREAK, True), ids=("BREAK", "True"))
    def allows_early_exit_from_leaving_different_points(break_action):
        ast = parse("{ a { y }, b { x } }")
        visited = []

        class TestVisitor(Visitor):
            def __init__(self, name):
                super().__init__()
                self.name = name

            def enter(self, *args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                name = self.name
                visited.append([f"break-{name}", "enter", kind, value])

            def leave(self, *args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                name = self.name
                visited.append([f"break-{name}", "leave", kind, value])
                if kind == "field" and node.name.value == name:
                    return break_action

        visit(ast, ParallelVisitor([TestVisitor("a"), TestVisitor("b")]))
        assert visited == [
            ["break-a", "enter", "document", None],
            ["break-b", "enter", "document", None],
            ["break-a", "enter", "operation_definition", None],
            ["break-b", "enter", "operation_definition", None],
            ["break-a", "enter", "selection_set", None],
            ["break-b", "enter", "selection_set", None],
            ["break-a", "enter", "field", None],
            ["break-b", "enter", "field", None],
            ["break-a", "enter", "name", "a"],
            ["break-b", "enter", "name", "a"],
            ["break-a", "leave", "name", "a"],
            ["break-b", "leave", "name", "a"],
            ["break-a", "enter", "selection_set", None],
            ["break-b", "enter", "selection_set", None],
            ["break-a", "enter", "field", None],
            ["break-b", "enter", "field", None],
            ["break-a", "enter", "name", "y"],
            ["break-b", "enter", "name", "y"],
            ["break-a", "leave", "name", "y"],
            ["break-b", "leave", "name", "y"],
            ["break-a", "leave", "field", None],
            ["break-b", "leave", "field", None],
            ["break-a", "leave", "selection_set", None],
            ["break-b", "leave", "selection_set", None],
            ["break-a", "leave", "field", None],
            ["break-b", "leave", "field", None],
            ["break-b", "enter", "field", None],
            ["break-b", "enter", "name", "b"],
            ["break-b", "leave", "name", "b"],
            ["break-b", "enter", "selection_set", None],
            ["break-b", "enter", "field", None],
            ["break-b", "enter", "name", "x"],
            ["break-b", "leave", "name", "x"],
            ["break-b", "leave", "field", None],
            ["break-b", "leave", "selection_set", None],
            ["break-b", "leave", "field", None],
        ]

    @mark.parametrize("remove_action", (REMOVE, Ellipsis), ids=("REMOVE", "Ellipsis"))
    def allows_for_editing_on_enter(remove_action):
        ast = parse("{ a, b, c { a, b, c } }", no_location=True)
        visited = []

        class TestVisitor1(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                if node.kind == "field" and node.name.value == "b":
                    return remove_action

        class TestVisitor2(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def leave(*args):
                check_visitor_fn_args_edited(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        edited_ast = visit(ast, ParallelVisitor([TestVisitor1(), TestVisitor2()]))
        assert ast == parse("{ a, b, c { a, b, c } }", no_location=True)
        assert edited_ast == parse("{ a,    c { a,    c } }", no_location=True)
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "c"],
            ["leave", "name", "c"],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "c"],
            ["leave", "name", "c"],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "operation_definition", None],
            ["leave", "document", None],
        ]

    @mark.parametrize("remove_action", (REMOVE, Ellipsis), ids=("REMOVE", "Ellipsis"))
    def allows_for_editing_on_leave(remove_action):
        ast = parse("{ a, b, c { a, b, c } }", no_location=True)
        visited = []

        class TestVisitor1(Visitor):
            @staticmethod
            def leave(*args):
                check_visitor_fn_args_edited(ast, *args)
                node = args[0]
                if node.kind == "field" and node.name.value == "b":
                    return remove_action

        class TestVisitor2(Visitor):
            @staticmethod
            def enter(*args):
                check_visitor_fn_args(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["enter", kind, value])

            @staticmethod
            def leave(*args):
                check_visitor_fn_args_edited(ast, *args)
                node = args[0]
                kind, value = node.kind, get_value(node)
                visited.append(["leave", kind, value])

        edited_ast = visit(ast, ParallelVisitor([TestVisitor1(), TestVisitor2()]))
        assert ast == parse("{ a, b, c { a, b, c } }", no_location=True)
        assert edited_ast == parse("{ a,    c { a,    c } }", no_location=True)
        assert visited == [
            ["enter", "document", None],
            ["enter", "operation_definition", None],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "b"],
            ["leave", "name", "b"],
            ["enter", "field", None],
            ["enter", "name", "c"],
            ["leave", "name", "c"],
            ["enter", "selection_set", None],
            ["enter", "field", None],
            ["enter", "name", "a"],
            ["leave", "name", "a"],
            ["leave", "field", None],
            ["enter", "field", None],
            ["enter", "name", "b"],
            ["leave", "name", "b"],
            ["enter", "field", None],
            ["enter", "name", "c"],
            ["leave", "name", "c"],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "field", None],
            ["leave", "selection_set", None],
            ["leave", "operation_definition", None],
            ["leave", "document", None],
        ]



from typing import Any, Callable, Collection, Dict, List, Optional, Union, cast

from ..error import GraphQLError
from ..language import (
    DirectiveNode,
    EnumValueDefinitionNode,
    ExecutableDefinitionNode,
    FieldNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
    NullValueNode,
    SchemaDefinitionNode,
    SelectionNode,
    TypeDefinitionNode,
    TypeExtensionNode,
    VariableDefinitionNode,
    VariableNode,
    print_ast,
)
from ..pyutils import inspect, print_path_list, Undefined
from ..type import (
    GraphQLDirective,
    GraphQLField,
    GraphQLInputType,
    GraphQLSchema,
    is_input_type,
    is_non_null_type,
)
from ..utilities.coerce_input_value import coerce_input_value
from ..utilities.type_from_ast import type_from_ast
from ..utilities.value_from_ast import value_from_ast

__all__ = ["get_argument_values", "get_directive_values", "get_variable_values"]


CoercedVariableValues = Union[List[GraphQLError], Dict[str, Any]]


def get_variable_values(
    schema: GraphQLSchema,
    var_def_nodes: Collection[VariableDefinitionNode],
    inputs: Dict[str, Any],
    max_errors: Optional[int] = None,
) -> CoercedVariableValues:
    """Get coerced variable values based on provided definitions.

    Prepares a dict of variable values of the correct type based on the provided
    variable definitions and arbitrary input. If the input cannot be parsed to match
    the variable definitions, a GraphQLError will be raised.
    """
    errors: List[GraphQLError] = []

    def on_error(error: GraphQLError) -> None:
        if max_errors is not None and len(errors) >= max_errors:
            raise GraphQLError(
                "Too many errors processing variables,"
                " error limit reached. Execution aborted."
            )
        errors.append(error)

    try:
        coerced = coerce_variable_values(schema, var_def_nodes, inputs, on_error)
        if not errors:
            return coerced
    except GraphQLError as e:
        errors.append(e)

    return errors


def coerce_variable_values(
    schema: GraphQLSchema,
    var_def_nodes: Collection[VariableDefinitionNode],
    inputs: Dict[str, Any],
    on_error: Callable[[GraphQLError], None],
) -> Dict[str, Any]:
    coerced_values: Dict[str, Any] = {}
    for var_def_node in var_def_nodes:
        var_name = var_def_node.variable.name.value
        var_type = type_from_ast(schema, var_def_node.type)
        if not is_input_type(var_type):
            # Must use input types for variables. This should be caught during
            # validation, however is checked again here for safety.
            var_type_str = print_ast(var_def_node.type)
            on_error(
                GraphQLError(
                    f"Variable '${var_name}' expected value of type '{var_type_str}'"
                    " which cannot be used as an input type.",
                    var_def_node.type,
                )
            )
            continue

        var_type = cast(GraphQLInputType, var_type)
        if var_name not in inputs:
            if var_def_node.default_value:
                coerced_values[var_name] = value_from_ast(
                    var_def_node.default_value, var_type
                )
            elif is_non_null_type(var_type):  # pragma: no cover else
                var_type_str = inspect(var_type)
                on_error(
                    GraphQLError(
                        f"Variable '${var_name}' of required type '{var_type_str}'"
                        " was not provided.",
                        var_def_node,
                    )
                )
            continue

        value = inputs[var_name]
        if value is None and is_non_null_type(var_type):
            var_type_str = inspect(var_type)
            on_error(
                GraphQLError(
                    f"Variable '${var_name}' of non-null type '{var_type_str}'"
                    " must not be null.",
                    var_def_node,
                )
            )
            continue

        def on_input_value_error(
            path: List[Union[str, int]], invalid_value: Any, error: GraphQLError
        ) -> None:
            invalid_str = inspect(invalid_value)
            prefix = f"Variable '${var_name}' got invalid value {invalid_str}"
            if path:
                prefix += f" at '{var_name}{print_path_list(path)}'"
            on_error(
                GraphQLError(
                    prefix + "; " + error.message,
                    var_def_node,
                    original_error=error.original_error,
                )
            )

        coerced_values[var_name] = coerce_input_value(
            value, var_type, on_input_value_error
        )

    return coerced_values


def get_argument_values(
    type_def: Union[GraphQLField, GraphQLDirective],
    node: Union[FieldNode, DirectiveNode],
    variable_values: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Get coerced argument values based on provided definitions and nodes.

    Prepares a dict of argument values given a list of argument definitions and list
    of argument AST nodes.
    """
    coerced_values: Dict[str, Any] = {}
    arg_node_map = {arg.name.value: arg for arg in node.arguments or []}

    for name, arg_def in type_def.args.items():
        arg_type = arg_def.type
        argument_node = arg_node_map.get(name)

        if argument_node is None:
            if arg_def.default_value is not Undefined:
                coerced_values[arg_def.out_name or name] = arg_def.default_value
            elif is_non_null_type(arg_type):  # pragma: no cover else
                raise GraphQLError(
                    f"Argument '{name}' of required type '{arg_type}'"
                    " was not provided.",
                    node,
                )
            continue  # pragma: no cover

        value_node = argument_node.value
        is_null = isinstance(argument_node.value, NullValueNode)

        if isinstance(value_node, VariableNode):
            variable_name = value_node.name.value
            if variable_values is None or variable_name not in variable_values:
                if arg_def.default_value is not Undefined:
                    coerced_values[arg_def.out_name or name] = arg_def.default_value
                elif is_non_null_type(arg_type):  # pragma: no cover else
                    raise GraphQLError(
                        f"Argument '{name}' of required type '{arg_type}'"
                        f" was provided the variable '${variable_name}'"
                        " which was not provided a runtime value.",
                        value_node,
                    )
                continue  # pragma: no cover
            is_null = variable_values[variable_name] is None

        if is_null and is_non_null_type(arg_type):
            raise GraphQLError(
                f"Argument '{name}' of non-null type '{arg_type}' must not be null.",
                value_node,
            )

        coerced_value = value_from_ast(value_node, arg_type, variable_values)
        if coerced_value is Undefined:
            # Note: `values_of_correct_type` validation should catch this before
            # execution. This is a runtime check to ensure execution does not
            # continue with an invalid argument value.
            raise GraphQLError(
                f"Argument '{name}' has invalid value {print_ast(value_node)}.",
                value_node,
            )
        coerced_values[arg_def.out_name or name] = coerced_value

    return coerced_values


NodeWithDirective = Union[
    EnumValueDefinitionNode,
    ExecutableDefinitionNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
    SelectionNode,
    SchemaDefinitionNode,
    TypeDefinitionNode,
    TypeExtensionNode,
]


def get_directive_values(
    directive_def: GraphQLDirective,
    node: NodeWithDirective,
    variable_values: Optional[Dict[str, Any]] = None,
) -> Optional[Dict[str, Any]]:
    """Get coerced argument values based on provided nodes.

    Prepares a dict of argument values given a directive definition and an AST node
    which may contain directives. Optionally also accepts a dict of variable values.

    If the directive does not exist on the node, returns None.
    """
    directives = node.directives
    if directives:
        directive_name = directive_def.name
        for directive in directives:
            if directive.name.value == directive_name:
                return get_argument_values(directive_def, directive, variable_values)
    return None



from typing import Any, Dict, List, Optional, cast

from ..language import (
    ListValueNode,
    NullValueNode,
    ObjectValueNode,
    ValueNode,
    VariableNode,
)
from ..pyutils import inspect, Undefined
from ..type import (
    GraphQLInputObjectType,
    GraphQLInputType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
    is_input_object_type,
    is_leaf_type,
    is_list_type,
    is_non_null_type,
)

__all__ = ["value_from_ast"]


def value_from_ast(
    value_node: Optional[ValueNode],
    type_: GraphQLInputType,
    variables: Optional[Dict[str, Any]] = None,
) -> Any:
    """Produce a Python value given a GraphQL Value AST.

    A GraphQL type must be provided, which will be used to interpret different GraphQL
    Value literals.

    Returns ``Undefined`` when the value could not be validly coerced according
    to the provided type.

    =================== ============== ================
       GraphQL Value      JSON Value     Python Value
    =================== ============== ================
       Input Object       Object         dict
       List               Array          list
       Boolean            Boolean        bool
       String             String         str
       Int / Float        Number         int / float
       Enum Value         Mixed          Any
       NullValue          null           None
    =================== ============== ================

    """
    if not value_node:
        # When there is no node, then there is also no value.
        # Importantly, this is different from returning the value null.
        return Undefined

    if isinstance(value_node, VariableNode):
        variable_name = value_node.name.value
        if not variables:
            return Undefined
        variable_value = variables.get(variable_name, Undefined)
        if variable_value is None and is_non_null_type(type_):
            return Undefined
        # Note: This does no further checking that this variable is correct.
        # This assumes that this query has been validated and the variable usage here
        # is of the correct type.
        return variable_value

    if is_non_null_type(type_):
        if isinstance(value_node, NullValueNode):
            return Undefined
        type_ = cast(GraphQLNonNull, type_)
        return value_from_ast(value_node, type_.of_type, variables)

    if isinstance(value_node, NullValueNode):
        return None  # This is explicitly returning the value None.

    if is_list_type(type_):
        type_ = cast(GraphQLList, type_)
        item_type = type_.of_type
        if isinstance(value_node, ListValueNode):
            coerced_values: List[Any] = []
            append_value = coerced_values.append
            for item_node in value_node.values:
                if is_missing_variable(item_node, variables):
                    # If an array contains a missing variable, it is either coerced to
                    # None or if the item type is non-null, it is considered invalid.
                    if is_non_null_type(item_type):
                        return Undefined
                    append_value(None)
                else:
                    item_value = value_from_ast(item_node, item_type, variables)
                    if item_value is Undefined:
                        return Undefined
                    append_value(item_value)
            return coerced_values
        coerced_value = value_from_ast(value_node, item_type, variables)
        if coerced_value is Undefined:
            return Undefined
        return [coerced_value]

    if is_input_object_type(type_):
        if not isinstance(value_node, ObjectValueNode):
            return Undefined
        type_ = cast(GraphQLInputObjectType, type_)
        coerced_obj: Dict[str, Any] = {}
        fields = type_.fields
        field_nodes = {field.name.value: field for field in value_node.fields}
        for field_name, field in fields.items():
            field_node = field_nodes.get(field_name)
            if not field_node or is_missing_variable(field_node.value, variables):
                if field.default_value is not Undefined:
                    # Use out name as name if it exists (extension of GraphQL.js).
                    coerced_obj[field.out_name or field_name] = field.default_value
                elif is_non_null_type(field.type):  # pragma: no cover else
                    return Undefined
                continue
            field_value = value_from_ast(field_node.value, field.type, variables)
            if field_value is Undefined:
                return Undefined
            coerced_obj[field.out_name or field_name] = field_value

        return type_.out_type(coerced_obj)

    if is_leaf_type(type_):
        # Scalars fulfill parsing a literal value via `parse_literal()`. Invalid values
        # represent a failure to parse correctly, in which case Undefined is returned.
        type_ = cast(GraphQLScalarType, type_)
        # noinspection PyBroadException
        try:
            if variables:
                result = type_.parse_literal(value_node, variables)
            else:
                result = type_.parse_literal(value_node)
        except Exception:
            return Undefined
        return result

    # Not reachable. All possible input types have been considered.
    raise TypeError(f"Unexpected input type: {inspect(type_)}.")


def is_missing_variable(
    value_node: ValueNode, variables: Optional[Dict[str, Any]] = None
) -> bool:
    """Check if ``value_node`` is a variable not defined in the ``variables`` dict."""
    return isinstance(value_node, VariableNode) and (
        not variables or variables.get(value_node.name.value, Undefined) is Undefined
    )



from graphql.pyutils import natural_comparison_key

key = natural_comparison_key


def describe_natural_compare():
    def handles_empty_strings():
        assert key("") < key("a")
        assert key("") < key("1")

    def handles_strings_of_different_length():
        assert key("A") < key("AA")
        assert key("A1") < key("A1A")

    def handles_numbers():
        assert key("1") < key("2")
        assert key("2") < key("11")

    def handles_numbers_with_leading_zeros():
        assert key("0") < key("00")
        assert key("02") < key("11")
        assert key("011") < key("200")

    def handles_numbers_embedded_into_names():
        assert key("a0a") < key("a9a")
        assert key("a00a") < key("a09a")
        assert key("a0a1") < key("a0a9")
        assert key("a10a11a") < key("a10a19a")
        assert key("a10a11a") < key("a10a11b")



from textwrap import dedent
from typing import List, Optional, Tuple

from pytest import raises

from graphql.error import GraphQLSyntaxError
from graphql.language import (
    ArgumentNode,
    BooleanValueNode,
    DirectiveDefinitionNode,
    DirectiveNode,
    DocumentNode,
    EnumTypeDefinitionNode,
    EnumValueDefinitionNode,
    FieldDefinitionNode,
    InputObjectTypeDefinitionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    InterfaceTypeExtensionNode,
    ListTypeNode,
    NameNode,
    NamedTypeNode,
    NonNullTypeNode,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
    OperationType,
    OperationTypeDefinitionNode,
    ScalarTypeDefinitionNode,
    SchemaDefinitionNode,
    SchemaExtensionNode,
    StringValueNode,
    TypeNode,
    UnionTypeDefinitionNode,
    ValueNode,
    parse,
)

from ..fixtures import kitchen_sink_sdl  # noqa: F401

Location = Optional[Tuple[int, int]]


def assert_syntax_error(text: str, message: str, location: Location) -> None:
    with raises(GraphQLSyntaxError) as exc_info:
        parse(text)
    error = exc_info.value
    assert error.message == f"Syntax Error: {message}"
    assert error.description == message
    assert error.locations == [location]


def assert_definitions(body: str, loc: Location, num=1):
    doc = parse(body)
    assert isinstance(doc, DocumentNode)
    assert doc.loc == loc
    definitions = doc.definitions
    assert isinstance(definitions, tuple)
    assert len(definitions) == num
    return definitions[0] if num == 1 else definitions


def type_node(name: str, loc: Location):
    return NamedTypeNode(name=name_node(name, loc), loc=loc)


def name_node(name: str, loc: Location):
    return NameNode(value=name, loc=loc)


def field_node(name: NameNode, type_: TypeNode, loc: Location):
    return field_node_with_args(name, type_, [], loc)


def field_node_with_args(name: NameNode, type_: TypeNode, args: List, loc: Location):
    return FieldDefinitionNode(
        name=name, arguments=args, type=type_, directives=[], loc=loc, description=None
    )


def non_null_type(type_: TypeNode, loc: Location):
    return NonNullTypeNode(type=type_, loc=loc)


def enum_value_node(name: str, loc: Location):
    return EnumValueDefinitionNode(
        name=name_node(name, loc), directives=[], loc=loc, description=None
    )


def input_value_node(
    name: NameNode, type_: TypeNode, default_value: Optional[ValueNode], loc: Location
):
    return InputValueDefinitionNode(
        name=name,
        type=type_,
        default_value=default_value,
        directives=[],
        loc=loc,
        description=None,
    )


def boolean_value_node(value: bool, loc: Location):
    return BooleanValueNode(value=value, loc=loc)


def string_value_node(value: str, block: Optional[bool], loc: Location):
    return StringValueNode(value=value, block=block, loc=loc)


def list_type_node(type_: TypeNode, loc: Location):
    return ListTypeNode(type=type_, loc=loc)


def schema_extension_node(
    directives: List[DirectiveNode],
    operation_types: List[OperationTypeDefinitionNode],
    loc: Location,
):
    return SchemaExtensionNode(
        directives=directives, operation_types=operation_types, loc=loc
    )


def operation_type_definition(operation: OperationType, type_: TypeNode, loc: Location):
    return OperationTypeDefinitionNode(operation=operation, type=type_, loc=loc)


def directive_node(name: NameNode, arguments: List[ArgumentNode], loc: Location):
    return DirectiveNode(name=name, arguments=arguments, loc=loc)


def describe_schema_parser():
    def simple_type():
        body = dedent(
            """
            type Hello {
              world: String
            }
            """
        )
        definition = assert_definitions(body, (0, 32))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("world", (16, 21)), type_node("String", (23, 29)), (16, 29)
            ),
        )
        assert definition.loc == (1, 31)

    def parses_type_with_description_string():
        body = dedent(
            """
            "Description"
            type Hello {
              world: String
            }
            """
        )
        definition = assert_definitions(body, (0, 46))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (20, 25))
        description = definition.description
        assert description == string_value_node("Description", False, (1, 14))

    def parses_type_with_description_multi_line_string():
        body = dedent(
            '''
            """
            Description
            """
            # Even with comments between them
            type Hello {
              world: String
            }'''
        )
        definition = assert_definitions(body, (0, 85))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (60, 65))
        description = definition.description
        assert description == string_value_node("Description", True, (1, 20))

    def parses_schema_with_description_string():
        body = dedent(
            """
            "Description"
            schema {
              query: Foo
            }
            """
        )
        definition = assert_definitions(body, (0, 39))
        assert isinstance(definition, SchemaDefinitionNode)
        description = definition.description
        assert description == string_value_node("Description", False, (1, 14))

    def description_followed_by_something_other_than_type_system_definition_throws():
        assert_syntax_error('"Description" 1', "Unexpected Int '1'.", (1, 15))

    def simple_extension():
        body = dedent(
            """
            extend type Hello {
              world: String
            }
            """
        )
        extension = assert_definitions(body, (0, 39))
        assert isinstance(extension, ObjectTypeExtensionNode)
        assert extension.name == name_node("Hello", (13, 18))
        assert extension.interfaces == ()
        assert extension.directives == ()
        assert extension.fields == (
            field_node(
                name_node("world", (23, 28)), type_node("String", (30, 36)), (23, 36)
            ),
        )
        assert extension.loc == (1, 38)

    def object_extension_without_fields():
        body = "extend type Hello implements Greeting"
        extension = assert_definitions(body, (0, 37))
        assert isinstance(extension, ObjectTypeExtensionNode)
        assert extension.name == name_node("Hello", (12, 17))
        assert extension.interfaces == (type_node("Greeting", (29, 37)),)
        assert extension.directives == ()
        assert extension.fields == ()
        assert extension.loc == (0, 37)

    def interface_extension_without_fields():
        body = "extend interface Hello implements Greeting"
        extension = assert_definitions(body, (0, 42))
        assert isinstance(extension, InterfaceTypeExtensionNode)
        assert extension.name == name_node("Hello", (17, 22))
        assert extension.interfaces == (type_node("Greeting", (34, 42)),)
        assert extension.directives == ()
        assert extension.fields == ()
        assert extension.loc == (0, 42)

    def object_extension_without_fields_followed_by_extension():
        body = (
            "\n      extend type Hello implements Greeting\n\n"
            "      extend type Hello implements SecondGreeting\n    "
        )
        extensions = assert_definitions(body, (0, 100), 2)
        extension = extensions[0]
        assert isinstance(extension, ObjectTypeExtensionNode)
        assert extension.name == name_node("Hello", (19, 24))
        assert extension.interfaces == (type_node("Greeting", (36, 44)),)
        assert extension.directives == ()
        assert extension.fields == ()
        assert extension.loc == (7, 44)
        extension = extensions[1]
        assert isinstance(extension, ObjectTypeExtensionNode)
        assert extension.name == name_node("Hello", (64, 69))
        assert extension.interfaces == (type_node("SecondGreeting", (81, 95)),)
        assert extension.directives == ()
        assert extension.fields == ()
        assert extension.loc == (52, 95)

    def extension_without_anything_throws():
        assert_syntax_error("extend scalar Hello", "Unexpected <EOF>.", (1, 20))
        assert_syntax_error("extend type Hello", "Unexpected <EOF>.", (1, 18))
        assert_syntax_error("extend interface Hello", "Unexpected <EOF>.", (1, 23))
        assert_syntax_error("extend union Hello", "Unexpected <EOF>.", (1, 19))
        assert_syntax_error("extend enum Hello", "Unexpected <EOF>.", (1, 18))
        assert_syntax_error("extend input Hello", "Unexpected <EOF>.", (1, 19))

    def interface_extension_without_fields_followed_by_extension():
        body = (
            "\n      extend interface Hello implements Greeting\n\n"
            "      extend interface Hello implements SecondGreeting\n    "
        )
        extensions = assert_definitions(body, (0, 110), 2)
        extension = extensions[0]
        assert isinstance(extension, InterfaceTypeExtensionNode)
        assert extension.name == name_node("Hello", (24, 29))
        assert extension.interfaces == (type_node("Greeting", (41, 49)),)
        assert extension.directives == ()
        assert extension.fields == ()
        assert extension.loc == (7, 49)
        extension = extensions[1]
        assert isinstance(extension, InterfaceTypeExtensionNode)
        assert extension.name == name_node("Hello", (74, 79))
        assert extension.interfaces == (type_node("SecondGreeting", (91, 105)),)
        assert extension.directives == ()
        assert extension.fields == ()
        assert extension.loc == (57, 105)

    def object_extension_do_not_include_descriptions():
        assert_syntax_error(
            """
            "Description"
            extend type Hello {
              world: String
            }""",
            "Unexpected description,"
            " descriptions are supported only on type definitions.",
            (2, 13),
        )
        assert_syntax_error(
            """
            extend "Description" type Hello {
              world: String
            }""",
            "Unexpected String 'Description'.",
            (2, 20),
        )

    def interface_extension_do_not_include_descriptions():
        assert_syntax_error(
            """
            "Description"
            extend interface Hello {
              world: String
            }""",
            "Unexpected description,"
            " descriptions are supported only on type definitions.",
            (2, 13),
        )
        assert_syntax_error(
            """
            extend "Description" interface Hello {
              world: String
            }""",
            "Unexpected String 'Description'.",
            (2, 20),
        )

    def schema_extension():
        body = """
            extend schema {
              mutation: Mutation
            }"""
        doc = parse(body)
        assert isinstance(doc, DocumentNode)
        assert doc.loc == (0, 75)
        assert doc.definitions == (
            schema_extension_node(
                [],
                [
                    operation_type_definition(
                        OperationType.MUTATION,
                        type_node("Mutation", (53, 61)),
                        (43, 61),
                    )
                ],
                (13, 75),
            ),
        )

    def schema_extension_with_only_directives():
        body = "extend schema @directive"
        doc = parse(body)
        assert isinstance(doc, DocumentNode)
        assert doc.loc == (0, 24)
        assert doc.definitions == (
            schema_extension_node(
                [directive_node(name_node("directive", (15, 24)), [], (14, 24))],
                [],
                (0, 24),
            ),
        )

    def schema_extension_without_anything_throws():
        assert_syntax_error("extend schema", "Unexpected <EOF>.", (1, 14))

    def schema_extension_with_invalid_operation_type_throws():
        assert_syntax_error(
            "extend schema { unknown: SomeType }", "Unexpected Name 'unknown'.", (1, 17)
        )

    def simple_non_null_type():
        body = dedent(
            """
            type Hello {
              world: String!
            }
            """
        )
        definition = assert_definitions(body, (0, 33))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("world", (16, 21)),
                non_null_type(type_node("String", (23, 29)), (23, 30)),
                (16, 30),
            ),
        )
        assert definition.loc == (1, 32)

    def simple_interface_inheriting_interface():
        body = "interface Hello implements World { field: String }"
        definition = assert_definitions(body, (0, 50))
        assert isinstance(definition, InterfaceTypeDefinitionNode)
        assert definition.name == name_node("Hello", (10, 15))
        assert definition.description is None
        assert definition.interfaces == (type_node("World", (27, 32)),)
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("field", (35, 40)), type_node("String", (42, 48)), (35, 48)
            ),
        )
        assert definition.loc == (0, 50)

    def simple_type_inheriting_interface():
        body = "type Hello implements World { field: String }"
        definition = assert_definitions(body, (0, 45))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (5, 10))
        assert definition.description is None
        assert definition.interfaces == (type_node("World", (22, 27)),)
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("field", (30, 35)), type_node("String", (37, 43)), (30, 43)
            ),
        )
        assert definition.loc == (0, 45)

    def simple_type_inheriting_multiple_interfaces():
        body = "type Hello implements Wo & rld { field: String }"
        definition = assert_definitions(body, (0, 48))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (5, 10))
        assert definition.description is None
        assert definition.interfaces == (
            type_node("Wo", (22, 24)),
            type_node("rld", (27, 30)),
        )
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("field", (33, 38)), type_node("String", (40, 46)), (33, 46)
            ),
        )
        assert definition.loc == (0, 48)

    def simple_interface_inheriting_multiple_interfaces():
        body = "interface Hello implements Wo & rld { field: String }"
        definition = assert_definitions(body, (0, 53))
        assert isinstance(definition, InterfaceTypeDefinitionNode)
        assert definition.name == name_node("Hello", (10, 15))
        assert definition.description is None
        assert definition.interfaces == (
            type_node("Wo", (27, 29)),
            type_node("rld", (32, 35)),
        )
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("field", (38, 43)), type_node("String", (45, 51)), (38, 51)
            ),
        )
        assert definition.loc == (0, 53)

    def simple_type_inheriting_multiple_interfaces_with_leading_ampersand():
        body = "type Hello implements & Wo & rld { field: String }"
        definition = assert_definitions(body, (0, 50))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (5, 10))
        assert definition.description is None
        assert definition.interfaces == (
            type_node("Wo", (24, 26)),
            type_node("rld", (29, 32)),
        )
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("field", (35, 40)), type_node("String", (42, 48)), (35, 48)
            ),
        )
        assert definition.loc == (0, 50)

    def simple_interface_inheriting_multiple_interfaces_with_leading_ampersand():
        body = "interface Hello implements & Wo & rld { field: String }"
        definition = assert_definitions(body, (0, 55))
        assert isinstance(definition, InterfaceTypeDefinitionNode)
        assert definition.name == name_node("Hello", (10, 15))
        assert definition.description is None
        assert definition.interfaces == (
            type_node("Wo", (29, 31)),
            type_node("rld", (34, 37)),
        )
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("field", (40, 45)), type_node("String", (47, 53)), (40, 53)
            ),
        )
        assert definition.loc == (0, 55)

    def single_value_enum():
        body = "enum Hello { WORLD }"
        definition = assert_definitions(body, (0, 20))
        assert isinstance(definition, EnumTypeDefinitionNode)
        assert definition.name == name_node("Hello", (5, 10))
        assert definition.description is None
        assert definition.directives == ()
        assert definition.values == (enum_value_node("WORLD", (13, 18)),)
        assert definition.loc == (0, 20)

    def double_value_enum():
        body = "enum Hello { WO, RLD }"
        definition = assert_definitions(body, (0, 22))
        assert isinstance(definition, EnumTypeDefinitionNode)
        assert definition.name == name_node("Hello", (5, 10))
        assert definition.description is None
        assert definition.directives == ()
        assert definition.values == (
            enum_value_node("WO", (13, 15)),
            enum_value_node("RLD", (17, 20)),
        )
        assert definition.loc == (0, 22)

    def simple_interface():
        body = dedent(
            """
            interface Hello {
              world: String
            }
            """
        )
        definition = assert_definitions(body, (0, 37))
        assert isinstance(definition, InterfaceTypeDefinitionNode)
        assert definition.name == name_node("Hello", (11, 16))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node(
                name_node("world", (21, 26)), type_node("String", (28, 34)), (21, 34)
            ),
        )
        assert definition.loc == (1, 36)

    def simple_field_with_arg():
        body = dedent(
            """
            type Hello {
              world(flag: Boolean): String
            }
            """
        )
        definition = assert_definitions(body, (0, 47))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node_with_args(
                name_node("world", (16, 21)),
                type_node("String", (38, 44)),
                [
                    input_value_node(
                        name_node("flag", (22, 26)),
                        type_node("Boolean", (28, 35)),
                        None,
                        (22, 35),
                    )
                ],
                (16, 44),
            ),
        )
        assert definition.loc == (1, 46)

    def simple_field_with_arg_with_default_value():
        body = dedent(
            """
            type Hello {
              world(flag: Boolean = true): String
            }
            """
        )
        definition = assert_definitions(body, (0, 54))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node_with_args(
                name_node("world", (16, 21)),
                type_node("String", (45, 51)),
                [
                    input_value_node(
                        name_node("flag", (22, 26)),
                        type_node("Boolean", (28, 35)),
                        boolean_value_node(True, (38, 42)),
                        (22, 42),
                    )
                ],
                (16, 51),
            ),
        )
        assert definition.loc == (1, 53)

    def simple_field_with_list_arg():
        body = dedent(
            """
            type Hello {
              world(things: [String]): String
            }
            """
        )
        definition = assert_definitions(body, (0, 50))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node_with_args(
                name_node("world", (16, 21)),
                type_node("String", (41, 47)),
                [
                    input_value_node(
                        name_node("things", (22, 28)),
                        list_type_node(type_node("String", (31, 37)), (30, 38)),
                        None,
                        (22, 38),
                    )
                ],
                (16, 47),
            ),
        )
        assert definition.loc == (1, 49)

    def simple_field_with_two_args():
        body = dedent(
            """
          type Hello {
            world(argOne: Boolean, argTwo: Int): String
          }
          """
        )
        definition = assert_definitions(body, (0, 62))
        assert isinstance(definition, ObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.interfaces == ()
        assert definition.directives == ()
        assert definition.fields == (
            field_node_with_args(
                name_node("world", (16, 21)),
                type_node("String", (53, 59)),
                [
                    input_value_node(
                        name_node("argOne", (22, 28)),
                        type_node("Boolean", (30, 37)),
                        None,
                        (22, 37),
                    ),
                    input_value_node(
                        name_node("argTwo", (39, 45)),
                        type_node("Int", (47, 50)),
                        None,
                        (39, 50),
                    ),
                ],
                (16, 59),
            ),
        )
        assert definition.loc == (1, 61)

    def simple_union():
        body = "union Hello = World"
        definition = assert_definitions(body, (0, 19))
        assert isinstance(definition, UnionTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.directives == ()
        assert definition.types == (type_node("World", (14, 19)),)
        assert definition.loc == (0, 19)

    def union_with_two_types():
        body = "union Hello = Wo | Rld"
        definition = assert_definitions(body, (0, 22))
        assert isinstance(definition, UnionTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.description is None
        assert definition.directives == ()
        assert definition.types == (
            type_node("Wo", (14, 16)),
            type_node("Rld", (19, 22)),
        )
        assert definition.loc == (0, 22)

    def union_with_two_types_and_leading_pipe():
        body = "union Hello = | Wo | Rld"
        definition = assert_definitions(body, (0, 24))
        assert isinstance(definition, UnionTypeDefinitionNode)
        assert definition.name == name_node("Hello", (6, 11))
        assert definition.directives == ()
        assert definition.types == (
            type_node("Wo", (16, 18)),
            type_node("Rld", (21, 24)),
        )
        assert definition.loc == (0, 24)

    def union_fails_with_no_types():
        assert_syntax_error("union Hello = |", "Expected Name, found <EOF>.", (1, 16))

    def union_fails_with_leading_double_pipe():
        assert_syntax_error(
            "union Hello = || Wo | Rld", "Expected Name, found '|'.", (1, 16)
        )

    def union_fails_with_double_pipe():
        assert_syntax_error(
            "union Hello = Wo || Rld", "Expected Name, found '|'.", (1, 19)
        )

    def union_fails_with_trailing_pipe():
        assert_syntax_error(
            "union Hello = | Wo | Rld |", "Expected Name, found <EOF>.", (1, 27)
        )

    def scalar():
        body = "scalar Hello"
        definition = assert_definitions(body, (0, 12))
        assert isinstance(definition, ScalarTypeDefinitionNode)
        assert definition.name == name_node("Hello", (7, 12))
        assert definition.description is None
        assert definition.directives == ()

        assert definition.loc == (0, 12)

    def simple_input_object():
        body = "\ninput Hello {\n  world: String\n}"
        definition = assert_definitions(body, (0, 32))
        assert isinstance(definition, InputObjectTypeDefinitionNode)
        assert definition.name == name_node("Hello", (7, 12))
        assert definition.description is None
        assert definition.directives == ()
        assert definition.fields == (
            input_value_node(
                name_node("world", (17, 22)),
                type_node("String", (24, 30)),
                None,
                (17, 30),
            ),
        )
        assert definition.loc == (1, 32)

    def simple_input_object_with_args_should_fail():
        assert_syntax_error(
            "\ninput Hello {\n  world(foo : Int): String\n}",
            "Expected ':', found '('.",
            (3, 8),
        )

    def directive_definition():
        body = "directive @foo on OBJECT | INTERFACE"
        definition = assert_definitions(body, (0, 36))
        assert isinstance(definition, DirectiveDefinitionNode)
        assert definition.name == name_node("foo", (11, 14))
        assert definition.description is None
        assert definition.arguments == ()
        assert definition.repeatable is False
        assert definition.locations == (
            name_node("OBJECT", (18, 24)),
            name_node("INTERFACE", (27, 36)),
        )

    def repeatable_directive_definition():
        body = "directive @foo repeatable on OBJECT | INTERFACE"
        definition = assert_definitions(body, (0, 47))
        assert isinstance(definition, DirectiveDefinitionNode)
        assert definition.name == name_node("foo", (11, 14))
        assert definition.description is None
        assert definition.arguments == ()
        assert definition.repeatable is True
        assert definition.locations == (
            name_node("OBJECT", (29, 35)),
            name_node("INTERFACE", (38, 47)),
        )

    def directive_with_incorrect_locations():
        assert_syntax_error(
            "\ndirective @foo on FIELD | INCORRECT_LOCATION",
            "Unexpected Name 'INCORRECT_LOCATION'.",
            (2, 27),
        )

    def parses_kitchen_sink_schema(kitchen_sink_sdl):  # noqa: F811
        assert parse(kitchen_sink_sdl)

    def can_pickle_and_unpickle_kitchen_sink_schema_ast(kitchen_sink_sdl):  # noqa: F811
        import pickle

        # create a schema AST from the kitchen sink SDL
        doc = parse(kitchen_sink_sdl)
        # check that the schema AST can be pickled
        # (particularly, there should be no recursion error)
        dumped = pickle.dumps(doc)
        # check that the pickle size is reasonable
        assert len(dumped) < 50 * len(kitchen_sink_sdl)
        loaded = pickle.loads(dumped)
        # check that the un-pickled schema AST is still the same
        assert loaded == doc
        # check that pickling again creates the same result
        dumped_again = pickle.dumps(doc)
        assert dumped_again == dumped



from typing import Any, Union, cast

from ...error import GraphQLError
from ...language import (
    DirectiveDefinitionNode,
    DocumentNode,
    ExecutableDefinitionNode,
    SchemaDefinitionNode,
    SchemaExtensionNode,
    TypeDefinitionNode,
    VisitorAction,
    SKIP,
)
from . import ASTValidationRule

__all__ = ["ExecutableDefinitionsRule"]


class ExecutableDefinitionsRule(ASTValidationRule):
    """Executable definitions

    A GraphQL document is only valid for execution if all definitions are either
    operation or fragment definitions.

    See https://spec.graphql.org/draft/#sec-Executable-Definitions
    """

    def enter_document(self, node: DocumentNode, *_args: Any) -> VisitorAction:
        for definition in node.definitions:
            if not isinstance(definition, ExecutableDefinitionNode):
                def_name = (
                    "schema"
                    if isinstance(
                        definition, (SchemaDefinitionNode, SchemaExtensionNode)
                    )
                    else "'{}'".format(
                        cast(
                            Union[DirectiveDefinitionNode, TypeDefinitionNode],
                            definition,
                        ).name.value
                    )
                )
                self.report_error(
                    GraphQLError(
                        f"The {def_name} definition is not executable.",
                        definition,
                    )
                )
        return SKIP



from graphql.language import parse, print_ast
from graphql.utilities import separate_operations

from ..utils import dedent


def separated_asts(ast):
    return {name: print_ast(node) for name, node in separate_operations(ast).items()}


def describe_separate_operations():
    def separates_one_ast_into_multiple_maintaining_document_order():
        ast = parse(
            """
            {
              ...Y
              ...X
            }

            query One {
              foo
              bar
              ...A
              ...X
            }

            fragment A on T {
              field
              ...B
            }

            fragment X on T {
              fieldX
            }

            query Two {
              ...A
              ...Y
              baz
            }

            fragment Y on T {
              fieldY
            }

            fragment B on T {
              something
            }

            """
        )

        assert separated_asts(ast) == {
            "": dedent(
                """
                {
                  ...Y
                  ...X
                }

                fragment X on T {
                  fieldX
                }

                fragment Y on T {
                  fieldY
                }
                """
            ),
            "One": dedent(
                """
                query One {
                  foo
                  bar
                  ...A
                  ...X
                }

                fragment A on T {
                  field
                  ...B
                }

                fragment X on T {
                  fieldX
                }

                fragment B on T {
                  something
                }
                """
            ),
            "Two": dedent(
                """
                fragment A on T {
                  field
                  ...B
                }

                query Two {
                  ...A
                  ...Y
                  baz
                }

                fragment Y on T {
                  fieldY
                }

                fragment B on T {
                  something
                }
                """
            ),
        }

    def survives_circular_dependencies():
        ast = parse(
            """
            query One {
              ...A
            }

            fragment A on T {
              ...B
            }

            fragment B on T {
              ...A
            }

            query Two {
              ...B
            }
            """
        )

        assert separated_asts(ast) == {
            "One": dedent(
                """
                query One {
                  ...A
                }

                fragment A on T {
                  ...B
                }

                fragment B on T {
                  ...A
                }
                """
            ),
            "Two": dedent(
                """
                fragment A on T {
                  ...B
                }

                fragment B on T {
                  ...A
                }

                query Two {
                  ...B
                }
                """
            ),
        }

    def distinguishes_query_and_fragment_names():
        ast = parse(
            """
            {
              ...NameClash
            }

            fragment NameClash on T {
              oneField
            }

            query NameClash {
              ...ShouldBeSkippedInFirstQuery
            }

            fragment ShouldBeSkippedInFirstQuery on T {
              twoField
            }
            """
        )

        assert separated_asts(ast) == {
            "": dedent(
                """
                {
                  ...NameClash
                }

                fragment NameClash on T {
                  oneField
                }
                """
            ),
            "NameClash": dedent(
                """
                query NameClash {
                  ...ShouldBeSkippedInFirstQuery
                }

                fragment ShouldBeSkippedInFirstQuery on T {
                  twoField
                }
                """
            ),
        }

    def handles_unknown_fragments():
        ast = parse(
            """
            {
              ...Unknown
              ...Known
            }

            fragment Known on T {
              someField
            }
            """
        )

        assert separated_asts(ast) == {
            "": dedent(
                """
                {
                  ...Unknown
                  ...Known
                }

                fragment Known on T {
                  someField
                }
                """
            )
        }



import re
from functools import partial
from typing import Any, Optional

from ...error import GraphQLError
from ...language import TypeDefinitionNode, TypeExtensionNode
from ...pyutils import did_you_mean, inspect, suggestion_list
from ...type import (
    is_enum_type,
    is_input_object_type,
    is_interface_type,
    is_object_type,
    is_scalar_type,
    is_union_type,
)
from . import SDLValidationContext, SDLValidationRule

__all__ = ["PossibleTypeExtensionsRule"]


class PossibleTypeExtensionsRule(SDLValidationRule):
    """Possible type extension

    A type extension is only valid if the type is defined and has the same kind.
    """

    def __init__(self, context: SDLValidationContext):
        super().__init__(context)
        self.schema = context.schema
        self.defined_types = {
            def_.name.value: def_
            for def_ in context.document.definitions
            if isinstance(def_, TypeDefinitionNode)
        }

    def check_extension(self, node: TypeExtensionNode, *_args: Any) -> None:
        schema = self.schema
        type_name = node.name.value
        def_node = self.defined_types.get(type_name)
        existing_type = schema.get_type(type_name) if schema else None

        expected_kind: Optional[str]
        if def_node:
            expected_kind = def_kind_to_ext_kind(def_node.kind)
        elif existing_type:
            expected_kind = type_to_ext_kind(existing_type)
        else:
            expected_kind = None

        if expected_kind:
            if expected_kind != node.kind:
                kind_str = extension_kind_to_type_name(node.kind)
                self.report_error(
                    GraphQLError(
                        f"Cannot extend non-{kind_str} type '{type_name}'.",
                        [def_node, node] if def_node else node,
                    )
                )
        else:
            all_type_names = list(self.defined_types)
            if self.schema:
                all_type_names.extend(self.schema.type_map)
            suggested_types = suggestion_list(type_name, all_type_names)
            self.report_error(
                GraphQLError(
                    f"Cannot extend type '{type_name}' because it is not defined."
                    + did_you_mean(suggested_types),
                    node.name,
                )
            )

    enter_scalar_type_extension = enter_object_type_extension = check_extension
    enter_interface_type_extension = enter_union_type_extension = check_extension
    enter_enum_type_extension = enter_input_object_type_extension = check_extension


def_kind_to_ext_kind = partial(re.compile("(?<=_type_)definition$").sub, "extension")


def type_to_ext_kind(type_: Any) -> str:
    if is_scalar_type(type_):
        return "scalar_type_extension"
    if is_object_type(type_):
        return "object_type_extension"
    if is_interface_type(type_):
        return "interface_type_extension"
    if is_union_type(type_):
        return "union_type_extension"
    if is_enum_type(type_):
        return "enum_type_extension"
    if is_input_object_type(type_):
        return "input_object_type_extension"

    # Not reachable. All possible types have been considered.
    raise TypeError(f"Unexpected type: {inspect(type_)}.")


_type_names_for_extension_kinds = {
    "scalar_type_extension": "scalar",
    "object_type_extension": "object",
    "interface_type_extension": "interface",
    "union_type_extension": "union",
    "enum_type_extension": "enum",
    "input_object_type_extension": "input object",
}


def extension_kind_to_type_name(kind: str) -> str:
    return _type_names_for_extension_kinds.get(kind, "unknown type")



from typing import Union

from ..error import GraphQLError
from ..language import (
    OperationType,
    OperationDefinitionNode,
    OperationTypeDefinitionNode,
)
from ..type import GraphQLObjectType, GraphQLSchema

__all__ = ["get_operation_root_type"]


def get_operation_root_type(
    schema: GraphQLSchema,
    operation: Union[OperationDefinitionNode, OperationTypeDefinitionNode],
) -> GraphQLObjectType:
    """Extract the root type of the operation from the schema.

    .. deprecated:: 3.2
       Please use `GraphQLSchema.getRootType` instead. Will be removed in v3.3.
    """
    operation_type = operation.operation
    if operation_type == OperationType.QUERY:
        query_type = schema.query_type
        if not query_type:
            raise GraphQLError(
                "Schema does not define the required query root type.", operation
            )
        return query_type

    if operation_type == OperationType.MUTATION:
        mutation_type = schema.mutation_type
        if not mutation_type:
            raise GraphQLError("Schema is not configured for mutations.", operation)
        return mutation_type

    if operation_type == OperationType.SUBSCRIPTION:
        subscription_type = schema.subscription_type
        if not subscription_type:
            raise GraphQLError("Schema is not configured for subscriptions.", operation)
        return subscription_type

    raise GraphQLError(
        "Can only have query, mutation and subscription operations.", operation
    )



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import KnownDirectivesRule

from .harness import assert_validation_errors, assert_sdl_validation_errors

schema_with_directives = build_schema(
    """
    type Query {
      dummy: String
    }

    directive @onQuery on QUERY
    directive @onMutation on MUTATION
    directive @onSubscription on SUBSCRIPTION
    directive @onField on FIELD
    directive @onFragmentDefinition on FRAGMENT_DEFINITION
    directive @onFragmentSpread on FRAGMENT_SPREAD
    directive @onInlineFragment on INLINE_FRAGMENT
    directive @onVariableDefinition on VARIABLE_DEFINITION
    """
)

schema_with_sdl_directives = build_schema(
    """
    directive @onSchema on SCHEMA
    directive @onScalar on SCALAR
    directive @onObject on OBJECT
    directive @onFieldDefinition on FIELD_DEFINITION
    directive @onArgumentDefinition on ARGUMENT_DEFINITION
    directive @onInterface on INTERFACE
    directive @onUnion on UNION
    directive @onEnum on ENUM
    directive @onEnumValue on ENUM_VALUE
    directive @onInputObject on INPUT_OBJECT
    directive @onInputFieldDefinition on INPUT_FIELD_DEFINITION
    """
)

assert_errors = partial(
    assert_validation_errors, KnownDirectivesRule, schema=schema_with_directives
)

assert_valid = partial(assert_errors, errors=[])

assert_sdl_errors = partial(assert_sdl_validation_errors, KnownDirectivesRule)

assert_sdl_valid = partial(assert_sdl_errors, errors=[])


def describe_known_directives():
    def with_no_directives():
        assert_valid(
            """
            query Foo {
              name
              ...Frag
            }

            fragment Frag on Dog {
              name
            }
            """
        )

    def with_standard_directives():
        assert_valid(
            """
            {
              human @skip(if: false) {
                name
                pets {
                  ... on Dog @include(if: true) {
                    name
                  }
                }
              }
            }
            """
        )

    def with_unknown_directive():
        assert_errors(
            """
            {
              human @unknown(directive: "value") {
                name
              }
            }
            """,
            [{"message": "Unknown directive '@unknown'.", "locations": [(3, 21)]}],
        )

    def with_many_unknown_directives():
        assert_errors(
            """
            {
              __typename @unknown
              human @unknown {
                name
                pets @unknown {
                  name
                }
              }
            }
            """,
            [
                {"message": "Unknown directive '@unknown'.", "locations": [(3, 26)]},
                {"message": "Unknown directive '@unknown'.", "locations": [(4, 21)]},
                {"message": "Unknown directive '@unknown'.", "locations": [(6, 22)]},
            ],
        )

    def with_well_placed_directives():
        assert_valid(
            """
            query ($var: Boolean @onVariableDefinition) @onQuery {
              human @onField {
                ...Frag @onFragmentSpread
                ... @onInlineFragment {
                  name @onField
                }
              }
            }

            mutation @onMutation {
              someField @onField
            }

            subscription @onSubscription {
              someField @onField
            }

            fragment Frag on Human @onFragmentDefinition {
              name @onField
            }
            """
        )

    def with_misplaced_directives():
        assert_errors(
            """
            query ($var: Boolean @onQuery) @onMutation {
              human @onQuery {
                ...Frag @onQuery
                ... @onQuery {
                  name @onQuery
                }
              }
            }

            mutation @onQuery {
              someField @onQuery
            }

            subscription @onQuery {
              someField @onQuery
            }

            fragment Frag on Human @onQuery {
              name @onQuery
            }
            """,
            [
                {
                    "message": "Directive '@onQuery'"
                    " may not be used on variable definition.",
                    "locations": [(2, 34)],
                },
                {
                    "message": "Directive '@onMutation' may not be used on query.",
                    "locations": [(2, 44)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on field.",
                    "locations": [(3, 21)],
                },
                {
                    "message": "Directive '@onQuery'"
                    " may not be used on fragment spread.",
                    "locations": [(4, 25)],
                },
                {
                    "message": "Directive '@onQuery'"
                    " may not be used on inline fragment.",
                    "locations": [(5, 21)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on field.",
                    "locations": [(6, 24)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on mutation.",
                    "locations": [(11, 22)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on field.",
                    "locations": [(12, 25)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on subscription.",
                    "locations": [(15, 26)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on field.",
                    "locations": [(16, 25)],
                },
                {
                    "message": "Directive '@onQuery'"
                    " may not be used on fragment definition.",
                    "locations": [(19, 36)],
                },
                {
                    "message": "Directive '@onQuery' may not be used on field.",
                    "locations": [(20, 20)],
                },
            ],
        )

    def with_misplaced_variable_definition_directive():
        assert_errors(
            """
            query Foo($var: Boolean @onField) {
              name
            }
            """,
            [
                {
                    "message": "Directive '@onField'"
                    " may not be used on variable definition.",
                    "locations": [(2, 37)],
                },
            ],
        )

    def describe_within_sdl():
        def with_directive_defined_inside_sdl():
            assert_sdl_valid(
                """
                type Query {
                  foo: String @test
                }

                directive @test on FIELD_DEFINITION
                """
            )

        def with_standard_directive():
            assert_sdl_valid(
                """
                type Query {
                  foo: String @deprecated
                }
                """
            )

        def with_overridden_standard_directive():
            assert_sdl_valid(
                """
                schema @deprecated {
                  query: Query
                }
                directive @deprecated on SCHEMA
                """
            )

        def with_directive_defined_in_schema_extension():
            schema = build_schema(
                """
                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_valid(
                """
                directive @test on OBJECT

                extend type Query @test
                """,
                schema=schema,
            )

        def with_directive_used_in_schema_extension():
            schema = build_schema(
                """
                directive @test on OBJECT

                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_valid(
                """
                extend type Query @test
                """,
                schema=schema,
            )

        def with_unknown_directive_in_schema_extension():
            schema = build_schema(
                """
                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_errors(
                """
                extend type Query @unknown
                """,
                [{"message": "Unknown directive '@unknown'.", "locations": [(2, 35)]}],
                schema,
            )

        def with_well_placed_directives():
            assert_sdl_valid(
                """
                type MyObj implements MyInterface @onObject {
                  myField(myArg: Int @onArgumentDefinition): String @onFieldDefinition
                }

                extend type MyObj @onObject

                scalar MyScalar @onScalar

                extend scalar MyScalar @onScalar

                interface MyInterface @onInterface {
                  myField(myArg: Int @onArgumentDefinition): String @onFieldDefinition
                }

                extend interface MyInterface @onInterface

                union MyUnion @onUnion = MyObj | Other

                extend union MyUnion @onUnion

                enum MyEnum @onEnum {
                  MY_VALUE @onEnumValue
                }

                extend enum MyEnum @onEnum

                input MyInput @onInputObject {
                  myField: Int @onInputFieldDefinition
                }

                extend input MyInput @onInputObject

                schema @onSchema {
                  query: MyQuery
                }

                extend schema @onSchema
                """,
                schema=schema_with_sdl_directives,
            )

        def with_misplaced_directives():
            assert_sdl_errors(
                """
                type MyObj implements MyInterface @onInterface {
                  myField(myArg: Int @onInputFieldDefinition): String @onInputFieldDefinition
                }

                scalar MyScalar @onEnum

                interface MyInterface @onObject {
                  myField(myArg: Int @onInputFieldDefinition): String @onInputFieldDefinition
                }

                union MyUnion @onEnumValue = MyObj | Other

                enum MyEnum @onScalar {
                  MY_VALUE @onUnion
                }

                input MyInput @onEnum {
                  myField: Int @onArgumentDefinition
                }

                schema @onObject {
                  query: MyQuery
                }

                extend schema @onObject
                """,  # noqa: E501
                [
                    {
                        "message": "Directive '@onInterface'"
                        " may not be used on object.",
                        "locations": [(2, 51)],
                    },
                    {
                        "message": "Directive '@onInputFieldDefinition'"
                        " may not be used on argument definition.",
                        "locations": [(3, 38)],
                    },
                    {
                        "message": "Directive '@onInputFieldDefinition'"
                        " may not be used on field definition.",
                        "locations": [(3, 71)],
                    },
                    {
                        "message": "Directive '@onEnum' may not be used on scalar.",
                        "locations": [(6, 33)],
                    },
                    {
                        "message": "Directive '@onObject'"
                        " may not be used on interface.",
                        "locations": [(8, 39)],
                    },
                    {
                        "message": "Directive '@onInputFieldDefinition'"
                        " may not be used on argument definition.",
                        "locations": [(9, 38)],
                    },
                    {
                        "message": "Directive '@onInputFieldDefinition'"
                        " may not be used on field definition.",
                        "locations": [(9, 71)],
                    },
                    {
                        "message": "Directive '@onEnumValue' may not be used on union.",
                        "locations": [(12, 31)],
                    },
                    {
                        "message": "Directive '@onScalar' may not be used on enum.",
                        "locations": [(14, 29)],
                    },
                    {
                        "message": "Directive '@onUnion'"
                        " may not be used on enum value.",
                        "locations": [(15, 28)],
                    },
                    {
                        "message": "Directive '@onEnum'"
                        " may not be used on input object.",
                        "locations": [(18, 31)],
                    },
                    {
                        "message": "Directive '@onArgumentDefinition'"
                        " may not be used on input field definition.",
                        "locations": [(19, 32)],
                    },
                    {
                        "message": "Directive '@onObject' may not be used on schema.",
                        "locations": [(22, 24)],
                    },
                    {
                        "message": "Directive '@onObject' may not be used on schema.",
                        "locations": [(26, 31)],
                    },
                ],
                schema_with_sdl_directives,
            )



from typing import Any, Dict, cast

from pytest import mark, param, raises

from graphql.type import (
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLSchema,
    GraphQLUnionType,
)

dummy_type = GraphQLScalarType("DummyScalar")

bad_extensions = [param([], id="list"), param({1: "ext"}, id="non_string_key")]


def bad_extensions_msg(name: str) -> str:
    return f"{name} extensions must be a dictionary with string keys."


def describe_type_system_extensions():
    def describe_graphql_scalar_type():
        def without_extensions():
            some_scalar = GraphQLScalarType("SomeScalar")
            assert some_scalar.extensions == {}
            assert some_scalar.to_kwargs()["extensions"] == {}

        def with_extensions():
            scalar_extensions = {"SomeScalarExt": "scalar"}
            some_scalar = GraphQLScalarType("SomeScalar", extensions=scalar_extensions)

            assert some_scalar.extensions is scalar_extensions
            assert some_scalar.to_kwargs()["extensions"] is scalar_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("SomeScalar")):
                # noinspection PyTypeChecker
                GraphQLScalarType("SomeScalar", extensions=extensions)

    def describe_graphql_object_type():
        def without_extensions():
            some_object = GraphQLObjectType(
                "SomeObject",
                {
                    "someField": GraphQLField(
                        dummy_type, {"someArg": GraphQLArgument(dummy_type)}
                    )
                },
            )

            assert some_object.extensions == {}
            some_field = some_object.fields["someField"]
            assert some_field.extensions == {}
            some_arg = some_field.args["someArg"]
            assert some_arg.extensions == {}

            assert some_object.to_kwargs()["extensions"] == {}
            assert some_field.to_kwargs()["extensions"] == {}
            assert some_arg.to_kwargs()["extensions"] == {}

        def with_extensions():
            object_extensions = {"SomeObjectExt": "object"}
            field_extensions = {"SomeFieldExt": "field"}
            arg_extensions = {"SomeArgExt": "arg"}

            some_object = GraphQLObjectType(
                "SomeObject",
                {
                    "someField": GraphQLField(
                        dummy_type,
                        {
                            "someArg": GraphQLArgument(
                                dummy_type, extensions=arg_extensions
                            )
                        },
                        extensions=field_extensions,
                    )
                },
                extensions=object_extensions,
            )

            assert some_object.extensions is object_extensions
            some_field = some_object.fields["someField"]
            assert some_field.extensions is field_extensions
            some_arg = some_field.args["someArg"]
            assert some_arg.extensions is arg_extensions

            assert some_object.to_kwargs()["extensions"] is object_extensions
            assert some_field.to_kwargs()["extensions"] is field_extensions
            assert some_arg.to_kwargs()["extensions"] is arg_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("SomeObject")):
                # noinspection PyTypeChecker
                GraphQLObjectType("SomeObject", {}, extensions=extensions)
            with raises(TypeError, match=bad_extensions_msg("Field")):
                # noinspection PyTypeChecker
                GraphQLField(dummy_type, extensions=extensions)
            with raises(TypeError, match=bad_extensions_msg("Argument")):
                # noinspection PyTypeChecker
                GraphQLArgument(dummy_type, extensions=extensions)

    def describe_graphql_interface_type():
        def without_extensions():
            some_interface = GraphQLInterfaceType(
                "SomeInterface",
                {
                    "someField": GraphQLField(
                        dummy_type, {"someArg": GraphQLArgument(dummy_type)}
                    )
                },
            )

            assert some_interface.extensions == {}
            some_field = some_interface.fields["someField"]
            assert some_field.extensions == {}
            some_arg = some_field.args["someArg"]
            assert some_arg.extensions == {}

            assert some_interface.to_kwargs()["extensions"] == {}
            assert some_field.to_kwargs()["extensions"] == {}
            assert some_arg.to_kwargs()["extensions"] == {}

        def with_extensions():
            interface_extensions = {"SomeInterfaceExt": "interface"}
            field_extensions = {"SomeFieldExt": "field"}
            arg_extensions = {"SomeArgExt": "arg"}

            some_interface = GraphQLInterfaceType(
                "SomeInterface",
                {
                    "someField": GraphQLField(
                        dummy_type,
                        {
                            "someArg": GraphQLArgument(
                                dummy_type, extensions=arg_extensions
                            )
                        },
                        extensions=field_extensions,
                    )
                },
                extensions=interface_extensions,
            )

            assert some_interface.extensions is interface_extensions
            some_field = some_interface.fields["someField"]
            assert some_field.extensions is field_extensions
            some_arg = some_field.args["someArg"]
            assert some_arg.extensions is arg_extensions

            assert some_interface.to_kwargs()["extensions"] is interface_extensions
            assert some_field.to_kwargs()["extensions"] is field_extensions
            assert some_arg.to_kwargs()["extensions"] is arg_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("SomeInterface")):
                # noinspection PyTypeChecker
                GraphQLInterfaceType("SomeInterface", {}, extensions=extensions)

    def describe_graphql_union_type():
        def without_extensions():
            some_union = GraphQLUnionType("SomeUnion", [])

            assert some_union.extensions == {}

            assert some_union.to_kwargs()["extensions"] == {}

        def with_extensions():
            union_extensions = {"SomeUnionExt": "union"}

            some_union = GraphQLUnionType("SomeUnion", [], extensions=union_extensions)

            assert some_union.extensions is union_extensions

            assert some_union.to_kwargs()["extensions"] is union_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("SomeUnion")):
                # noinspection PyTypeChecker
                GraphQLUnionType("SomeUnion", [], extensions=extensions)

    def describe_graphql_enum_type():
        def without_extensions():
            some_enum = GraphQLEnumType("SomeEnum", {"SOME_VALUE": None})

            assert some_enum.extensions == {}
            some_value = some_enum.values["SOME_VALUE"]
            assert some_value.extensions == {}

            assert some_enum.to_kwargs()["extensions"] == {}
            assert some_value.to_kwargs()["extensions"] == {}

        def with_extensions():
            enum_extensions = {"SomeEnumExt": "enum"}
            value_extensions = {"SomeValueExt": "value"}

            some_enum = GraphQLEnumType(
                "SomeEnum",
                {"SOME_VALUE": GraphQLEnumValue(extensions=value_extensions)},
                extensions=enum_extensions,
            )

            assert some_enum.extensions is enum_extensions
            some_value = some_enum.values["SOME_VALUE"]
            assert some_value.extensions is value_extensions

            assert some_enum.to_kwargs()["extensions"] is enum_extensions
            assert some_value.to_kwargs()["extensions"] is value_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("SomeEnum")):
                # noinspection PyTypeChecker
                GraphQLEnumType(
                    "SomeEnum", cast(Dict[str, Any], {}), extensions=extensions
                )
            with raises(TypeError, match=bad_extensions_msg("Enum value")):
                # noinspection PyTypeChecker
                GraphQLEnumValue(extensions=extensions)

    def describe_graphql_input_object_type():
        def without_extensions():
            some_input_object = GraphQLInputObjectType(
                "SomeInputObject", {"someInputField": GraphQLInputField(dummy_type)}
            )

            assert some_input_object.extensions == {}
            some_input_field = some_input_object.fields["someInputField"]
            assert some_input_field.extensions == {}

            assert some_input_object.to_kwargs()["extensions"] == {}
            assert some_input_field.to_kwargs()["extensions"] == {}

        def with_extensions():
            input_object_extensions = {"SomeInputObjectExt": "inputObject"}
            input_field_extensions = {"SomeInputFieldExt": "inputField"}

            some_input_object = GraphQLInputObjectType(
                "SomeInputObject",
                {
                    "someInputField": GraphQLInputField(
                        dummy_type, extensions=input_field_extensions
                    )
                },
                extensions=input_object_extensions,
            )

            assert some_input_object.extensions is input_object_extensions
            some_input_field = some_input_object.fields["someInputField"]
            assert some_input_field.extensions is input_field_extensions

            assert (
                some_input_object.to_kwargs()["extensions"] is input_object_extensions
            )
            assert some_input_field.to_kwargs()["extensions"] is input_field_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("SomeInputObject")):
                # noinspection PyTypeChecker
                GraphQLInputObjectType("SomeInputObject", {}, extensions=extensions)
            with raises(TypeError, match=bad_extensions_msg("Input field")):
                # noinspection PyTypeChecker
                GraphQLInputField(dummy_type, extensions=extensions)

    def describe_graphql_directive():
        def without_extensions():
            some_directive = GraphQLDirective(
                "SomeDirective", [], {"someArg": GraphQLArgument(dummy_type)}
            )

            assert some_directive.extensions == {}
            some_arg = some_directive.args["someArg"]
            assert some_arg.extensions == {}

            assert some_directive.to_kwargs()["extensions"] == {}
            assert some_arg.to_kwargs()["extensions"] == {}

        def with_extensions():
            directive_extensions = {"SomeDirectiveExt": "directive"}
            arg_extensions = {"SomeArgExt": "arg"}

            some_directive = GraphQLDirective(
                "SomeDirective",
                [],
                {"someArg": GraphQLArgument(dummy_type, extensions=arg_extensions)},
                extensions=directive_extensions,
            )

            assert some_directive.extensions is directive_extensions
            some_arg = some_directive.args["someArg"]
            assert some_arg.extensions is arg_extensions

            assert some_directive.to_kwargs()["extensions"] is directive_extensions
            assert some_arg.to_kwargs()["extensions"] is arg_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("Directive")):
                # noinspection PyTypeChecker
                GraphQLDirective("SomeDirective", [], extensions=extensions)

    def describe_graphql_schema():
        def without_extensions():
            schema = GraphQLSchema()

            assert schema.extensions == {}
            assert schema.to_kwargs()["extensions"] == {}

        def with_extensions():
            schema_extensions = {"schemaExtension": "schema"}

            schema = GraphQLSchema(extensions=schema_extensions)

            assert schema.extensions is schema_extensions

            assert schema.to_kwargs()["extensions"] is schema_extensions

        @mark.parametrize("extensions", bad_extensions)
        def with_bad_extensions(extensions):
            with raises(TypeError, match=bad_extensions_msg("Schema")):
                # noinspection PyTypeChecker
                GraphQLSchema(extensions=extensions)



from graphql.language import parse
from graphql.utilities import get_operation_ast


def describe_get_operation_ast():
    def gets_an_operation_from_a_simple_document():
        doc = parse("{ field }")
        assert get_operation_ast(doc) == doc.definitions[0]

    def gets_an_operation_from_a_document_with_named_op_mutation():
        doc = parse("mutation Test { field }")
        assert get_operation_ast(doc) == doc.definitions[0]

    def gets_an_operation_from_a_document_with_named_op_subscription():
        doc = parse("subscription Test { field }")
        assert get_operation_ast(doc) == doc.definitions[0]

    def does_not_get_missing_operation():
        doc = parse("type Foo { field: String }")
        assert get_operation_ast(doc) is None

    def does_not_get_ambiguous_unnamed_operation():
        doc = parse(
            """
            { field }
            mutation Test { field }
            subscription TestSub { field }
            """
        )
        assert get_operation_ast(doc) is None

    def does_not_get_ambiguous_named_operation():
        doc = parse(
            """
            query TestQ { field }
            mutation TestM { field }
            subscription TestS { field }
            """
        )
        assert get_operation_ast(doc) is None

    def does_not_get_misnamed_operation():
        doc = parse(
            """
            query TestQ { field }
            mutation TestM { field }
            subscription TestS { field }
            """
        )
        assert get_operation_ast(doc, "Unknown") is None

    def gets_named_operation():
        doc = parse(
            """
            query TestQ { field }
            mutation TestM { field }
            subscription TestS { field }
            """
        )
        assert get_operation_ast(doc, "TestQ") == doc.definitions[0]
        assert get_operation_ast(doc, "TestM") == doc.definitions[1]
        assert get_operation_ast(doc, "TestS") == doc.definitions[2]



from typing import Awaitable

from pytest import mark, raises

from graphql.execution import MiddlewareManager, execute
from graphql.language.parser import parse
from graphql.type import GraphQLField, GraphQLObjectType, GraphQLSchema, GraphQLString


def describe_middleware():
    def describe_with_manager():
        def default():
            doc = parse("{ field }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def field(self, _info):
                    return "resolved"

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            middlewares = MiddlewareManager()
            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )

            assert result.data["field"] == "resolved"  # type: ignore

        def single_function():
            doc = parse("{ first second }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def first(self, _info):
                    return "one"

                def second(self, _info):
                    return "two"

            test_type = GraphQLObjectType(
                "TestType",
                {
                    "first": GraphQLField(GraphQLString),
                    "second": GraphQLField(GraphQLString),
                },
            )

            def reverse_middleware(next_, *args, **kwargs):
                return next_(*args, **kwargs)[::-1]

            middlewares = MiddlewareManager(reverse_middleware)
            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )

            assert result.data == {"first": "eno", "second": "owt"}  # type: ignore

        def two_functions_and_field_resolvers():
            doc = parse("{ first second }")

            # noinspection PyMethodMayBeStatic
            class Data:
                first = "one"
                second = "two"

            test_type = GraphQLObjectType(
                "TestType",
                {
                    "first": GraphQLField(
                        GraphQLString, resolve=lambda obj, _info: obj.first
                    ),
                    "second": GraphQLField(
                        GraphQLString, resolve=lambda obj, _info: obj.second
                    ),
                },
            )

            def reverse_middleware(next_, *args, **kwargs):
                return next_(*args, **kwargs)[::-1]

            def capitalize_middleware(next_, *args, **kwargs):
                return next_(*args, **kwargs).capitalize()

            middlewares = MiddlewareManager(reverse_middleware, capitalize_middleware)
            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )

            assert result.data == {"first": "Eno", "second": "Owt"}  # type: ignore

        @mark.asyncio
        async def single_async_function():
            doc = parse("{ first second }")

            # noinspection PyMethodMayBeStatic
            class Data:
                async def first(self, _info):
                    return "one"

                async def second(self, _info):
                    return "two"

            test_type = GraphQLObjectType(
                "TestType",
                {
                    "first": GraphQLField(GraphQLString),
                    "second": GraphQLField(GraphQLString),
                },
            )

            async def reverse_middleware(next_, *args, **kwargs):
                return (await next_(*args, **kwargs))[::-1]

            middlewares = MiddlewareManager(reverse_middleware)
            awaitable_result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )
            assert isinstance(awaitable_result, Awaitable)
            result = await awaitable_result
            assert result.data == {"first": "eno", "second": "owt"}

        def single_object():
            doc = parse("{ first second }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def first(self, _info):
                    return "one"

                def second(self, _info):
                    return "two"

            test_type = GraphQLObjectType(
                "TestType",
                {
                    "first": GraphQLField(GraphQLString),
                    "second": GraphQLField(GraphQLString),
                },
            )

            class ReverseMiddleware:

                # noinspection PyMethodMayBeStatic
                def resolve(self, next_, *args, **kwargs):
                    return next_(*args, **kwargs)[::-1]

            middlewares = MiddlewareManager(ReverseMiddleware())
            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )

            assert result.data == {"first": "eno", "second": "owt"}  # type: ignore

        def skip_middleware_without_resolve_method():
            class BadMiddleware:
                pass  # no resolve method here

            assert execute(
                GraphQLSchema(
                    GraphQLObjectType(
                        "TestType",
                        {"foo": GraphQLField(GraphQLString)},
                    )
                ),
                parse("{ foo }"),
                {"foo": "bar"},
                middleware=MiddlewareManager(BadMiddleware()),
            ) == ({"foo": "bar"}, None)

        def with_function_and_object():
            doc = parse("{ field }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def field(self, _info):
                    return "resolved"

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            def reverse_middleware(next_, *args, **kwargs):
                return next_(*args, **kwargs)[::-1]

            class CaptitalizeMiddleware:

                # noinspection PyMethodMayBeStatic
                def resolve(self, next_, *args, **kwargs):
                    return next_(*args, **kwargs).capitalize()

            middlewares = MiddlewareManager(reverse_middleware, CaptitalizeMiddleware())
            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )
            assert result.data == {"field": "Devloser"}  # type: ignore

            middlewares = MiddlewareManager(CaptitalizeMiddleware(), reverse_middleware)
            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )
            assert result.data == {"field": "devloseR"}  # type: ignore

        @mark.asyncio
        async def with_async_function_and_object():
            doc = parse("{ field }")

            # noinspection PyMethodMayBeStatic
            class Data:
                async def field(self, _info):
                    return "resolved"

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            async def reverse_middleware(next_, *args, **kwargs):
                return (await next_(*args, **kwargs))[::-1]

            class CaptitalizeMiddleware:

                # noinspection PyMethodMayBeStatic
                async def resolve(self, next_, *args, **kwargs):
                    return (await next_(*args, **kwargs)).capitalize()

            middlewares = MiddlewareManager(reverse_middleware, CaptitalizeMiddleware())
            awaitable_result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )
            assert isinstance(awaitable_result, Awaitable)
            result = await awaitable_result
            assert result.data == {"field": "Devloser"}

            middlewares = MiddlewareManager(CaptitalizeMiddleware(), reverse_middleware)
            awaitable_result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )
            assert isinstance(awaitable_result, Awaitable)
            result = await awaitable_result
            assert result.data == {"field": "devloseR"}

    def describe_without_manager():
        def no_middleware():
            doc = parse("{ field }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def field(self, _info):
                    return "resolved"

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            result = execute(GraphQLSchema(test_type), doc, Data(), middleware=None)

            assert result.data["field"] == "resolved"  # type: ignore

        def empty_middleware_list():
            doc = parse("{ field }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def field(self, _info):
                    return "resolved"

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            result = execute(GraphQLSchema(test_type), doc, Data(), middleware=[])

            assert result.data["field"] == "resolved"  # type: ignore

        def bad_middleware_object():
            doc = parse("{ field }")

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            with raises(TypeError) as exc_info:
                # noinspection PyTypeChecker
                execute(
                    GraphQLSchema(test_type),
                    doc,
                    None,
                    middleware={"bad": "value"},  # type: ignore
                )

            assert str(exc_info.value) == (
                "Middleware must be passed as a list or tuple of functions"
                " or objects, or as a single MiddlewareManager object."
                " Got {'bad': 'value'} instead."
            )

        def list_of_functions():
            doc = parse("{ field }")

            # noinspection PyMethodMayBeStatic
            class Data:
                def field(self, _info):
                    return "resolved"

            test_type = GraphQLObjectType(
                "TestType", {"field": GraphQLField(GraphQLString)}
            )

            log = []

            class LogMiddleware:
                def __init__(self, name):
                    self.name = name

                # noinspection PyMethodMayBeStatic
                def resolve(self, next_, *args, **kwargs):
                    log.append(f"enter {self.name}")
                    value = next_(*args, **kwargs)
                    log.append(f"exit {self.name}")
                    return value

            middlewares = [LogMiddleware("A"), LogMiddleware("B"), LogMiddleware("C")]

            result = execute(
                GraphQLSchema(test_type), doc, Data(), middleware=middlewares
            )
            assert result.data == {"field": "resolved"}  # type: ignore

            assert log == [
                "enter C",
                "enter B",
                "enter A",
                "exit A",
                "exit B",
                "exit C",
            ]



from asyncio import ensure_future
from inspect import isawaitable
from typing import Any, Awaitable, Callable, Dict, Optional, Union, Type, cast

from .error import GraphQLError
from .execution import execute, ExecutionResult, ExecutionContext, Middleware
from .language import parse, Source
from .pyutils import AwaitableOrValue
from .type import (
    GraphQLFieldResolver,
    GraphQLSchema,
    GraphQLTypeResolver,
    validate_schema,
)

__all__ = ["graphql", "graphql_sync"]


async def graphql(
    schema: GraphQLSchema,
    source: Union[str, Source],
    root_value: Any = None,
    context_value: Any = None,
    variable_values: Optional[Dict[str, Any]] = None,
    operation_name: Optional[str] = None,
    field_resolver: Optional[GraphQLFieldResolver] = None,
    type_resolver: Optional[GraphQLTypeResolver] = None,
    middleware: Optional[Middleware] = None,
    execution_context_class: Optional[Type[ExecutionContext]] = None,
    is_awaitable: Optional[Callable[[Any], bool]] = None,
) -> ExecutionResult:
    """Execute a GraphQL operation asynchronously.

    This is the primary entry point function for fulfilling GraphQL operations by
    parsing, validating, and executing a GraphQL document along side a GraphQL schema.

    More sophisticated GraphQL servers, such as those which persist queries, may wish
    to separate the validation and execution phases to a static time tooling step,
    and a server runtime step.

    Accepts the following arguments:

    :arg schema:
      The GraphQL type system to use when validating and executing a query.
    :arg source:
      A GraphQL language formatted string representing the requested operation.
    :arg root_value:
      The value provided as the first argument to resolver functions on the top level
      type (e.g. the query object type).
    :arg context_value:
      The context value is provided as an attribute of the second argument
      (the resolve info) to resolver functions. It is used to pass shared information
      useful at any point during query execution, for example the currently logged in
      user and connections to databases or other services.
    :arg variable_values:
      A mapping of variable name to runtime value to use for all variables defined
      in the request string.
    :arg operation_name:
      The name of the operation to use if request string contains multiple possible
      operations. Can be omitted if request string contains only one operation.
    :arg field_resolver:
      A resolver function to use when one is not provided by the schema.
      If not provided, the default field resolver is used (which looks for a value
      or method on the source value with the field's name).
    :arg type_resolver:
      A type resolver function to use when none is provided by the schema.
      If not provided, the default type resolver is used (which looks for a
      ``__typename`` field or alternatively calls the
      :meth:`~graphql.type.GraphQLObjectType.is_type_of` method).
    :arg middleware:
      The middleware to wrap the resolvers with
    :arg execution_context_class:
      The execution context class to use to build the context
    :arg is_awaitable:
      The predicate to be used for checking whether values are awaitable
    """
    # Always return asynchronously for a consistent API.
    result = graphql_impl(
        schema,
        source,
        root_value,
        context_value,
        variable_values,
        operation_name,
        field_resolver,
        type_resolver,
        middleware,
        execution_context_class,
        is_awaitable,
    )

    if isawaitable(result):
        return await cast(Awaitable[ExecutionResult], result)

    return cast(ExecutionResult, result)


def assume_not_awaitable(_value: Any) -> bool:
    """Replacement for isawaitable if everything is assumed to be synchronous."""
    return False


def graphql_sync(
    schema: GraphQLSchema,
    source: Union[str, Source],
    root_value: Any = None,
    context_value: Any = None,
    variable_values: Optional[Dict[str, Any]] = None,
    operation_name: Optional[str] = None,
    field_resolver: Optional[GraphQLFieldResolver] = None,
    type_resolver: Optional[GraphQLTypeResolver] = None,
    middleware: Optional[Middleware] = None,
    execution_context_class: Optional[Type[ExecutionContext]] = None,
    check_sync: bool = False,
) -> ExecutionResult:
    """Execute a GraphQL operation synchronously.

    The graphql_sync function also fulfills GraphQL operations by parsing, validating,
    and executing a GraphQL document along side a GraphQL schema. However, it guarantees
    to complete synchronously (or throw an error) assuming that all field resolvers
    are also synchronous.

    Set check_sync to True to still run checks that no awaitable values are returned.
    """
    is_awaitable = (
        check_sync
        if callable(check_sync)
        else (None if check_sync else assume_not_awaitable)
    )
    result = graphql_impl(
        schema,
        source,
        root_value,
        context_value,
        variable_values,
        operation_name,
        field_resolver,
        type_resolver,
        middleware,
        execution_context_class,
        is_awaitable,
    )

    # Assert that the execution was synchronous.
    if isawaitable(result):
        ensure_future(cast(Awaitable[ExecutionResult], result)).cancel()
        raise RuntimeError("GraphQL execution failed to complete synchronously.")

    return cast(ExecutionResult, result)


def graphql_impl(
    schema: GraphQLSchema,
    source: Union[str, Source],
    root_value: Any,
    context_value: Any,
    variable_values: Optional[Dict[str, Any]],
    operation_name: Optional[str],
    field_resolver: Optional[GraphQLFieldResolver],
    type_resolver: Optional[GraphQLTypeResolver],
    middleware: Optional[Middleware],
    execution_context_class: Optional[Type[ExecutionContext]],
    is_awaitable: Optional[Callable[[Any], bool]],
) -> AwaitableOrValue[ExecutionResult]:
    """Execute a query, return asynchronously only if necessary."""
    # Validate Schema
    schema_validation_errors = validate_schema(schema)
    if schema_validation_errors:
        return ExecutionResult(data=None, errors=schema_validation_errors)

    # Parse
    try:
        document = parse(source)
    except GraphQLError as error:
        return ExecutionResult(data=None, errors=[error])

    # Validate
    from .validation import validate

    validation_errors = validate(schema, document)
    if validation_errors:
        return ExecutionResult(data=None, errors=validation_errors)

    # Execute
    return execute(
        schema,
        document,
        root_value,
        context_value,
        variable_values,
        operation_name,
        field_resolver,
        type_resolver,
        None,
        middleware,
        execution_context_class,
        is_awaitable,
    )



from functools import partial

from graphql.validation import NoUnusedFragmentsRule

from .harness import assert_validation_errors

assert_errors = partial(assert_validation_errors, NoUnusedFragmentsRule)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_no_unused_fragments():
    def all_fragment_names_are_used():
        assert_valid(
            """
            {
              human(id: 4) {
                ...HumanFields1
                ... on Human {
                  ...HumanFields2
                }
              }
            }
            fragment HumanFields1 on Human {
              name
              ...HumanFields3
            }
            fragment HumanFields2 on Human {
              name
            }
            fragment HumanFields3 on Human {
              name
            }
            """
        )

    def all_fragment_names_are_used_by_multiple_operations():
        assert_valid(
            """
            query Foo {
              human(id: 4) {
                ...HumanFields1
              }
            }
            query Bar {
              human(id: 4) {
                ...HumanFields2
              }
            }
            fragment HumanFields1 on Human {
              name
              ...HumanFields3
            }
            fragment HumanFields2 on Human {
              name
            }
            fragment HumanFields3 on Human {
              name
            }
            """
        )

    def contains_unknown_fragments():
        assert_errors(
            """
            query Foo {
              human(id: 4) {
                ...HumanFields1
              }
            }
            query Bar {
              human(id: 4) {
                ...HumanFields2
              }
            }
            fragment HumanFields1 on Human {
              name
              ...HumanFields3
            }
            fragment HumanFields2 on Human {
              name
            }
            fragment HumanFields3 on Human {
              name
            }
            fragment Unused1 on Human {
              name
            }
            fragment Unused2 on Human {
              name
            }
            """,
            [
                {
                    "message": "Fragment 'Unused1' is never used.",
                    "locations": [(22, 13)],
                },
                {
                    "message": "Fragment 'Unused2' is never used.",
                    "locations": [(25, 13)],
                },
            ],
        )

    def contains_unknown_fragments_with_ref_cycle():
        assert_errors(
            """
            query Foo {
              human(id: 4) {
                ...HumanFields1
              }
            }
            query Bar {
              human(id: 4) {
                ...HumanFields2
              }
            }
            fragment HumanFields1 on Human {
              name
              ...HumanFields3
            }
            fragment HumanFields2 on Human {
              name
            }
            fragment HumanFields3 on Human {
              name
            }
            fragment Unused1 on Human {
              name
              ...Unused2
            }
            fragment Unused2 on Human {
              name
              ...Unused1
            }
            """,
            [
                {
                    "message": "Fragment 'Unused1' is never used.",
                    "locations": [(22, 13)],
                },
                {
                    "message": "Fragment 'Unused2' is never used.",
                    "locations": [(26, 13)],
                },
            ],
        )

    def contains_unknown_and_undefined_fragments():
        assert_errors(
            """
            query Foo {
              human(id: 4) {
                ...bar
              }
            }
            fragment foo on Human {
              name
            }
            """,
            [{"message": "Fragment 'foo' is never used.", "locations": [(7, 13)]}],
        )
